import logging
import math
import operator
import sys
import tkinter as tk
from tkinter import ttk

logger = logging.getLogger(__name__)

# TODO: refactor IntegerEntry and NumericEntry to have common base class


class IntegerEntry(ttk.Entry):
    """Entry Entry that holds an integer value. Implements validation."""

    def __init__(
        self,
        parent,
        scroll_rate=1,
        min_value=float("-inf"),
        max_value=float("inf"),
        include_min_value=True,
        include_max_value=True,
        **kwargs,
    ):
        self.scroll_rate = scroll_rate
        self.min_value = min_value
        self.max_value = max_value
        self.include_min_value = include_min_value
        self.include_max_value = include_max_value
        kwargs["textvariable"] = kwargs.get("textvariable", None) or tk.StringVar()
        kwargs["validate"] = kwargs.get("validate", None) or "all"
        if "validatecommand" not in kwargs:
            validatecommand = (
                parent.register(self._validate),
                "%S",
                "%P",
                "%s",
                "%V",
                "%d",
            )
            kwargs["validatecommand"] = validatecommand
        if "invalidcommand" not in kwargs:
            invalidcommand = (
                parent.register(self._on_invalid),
                "%S",
                "%P",
                "%s",
                "%V",
                "%d",
            )
            kwargs["invalidcommand"] = invalidcommand
        self.var = kwargs["textvariable"]

        self._previous = ""  # used to recover after invalid "select all + paste"

        super().__init__(parent, **kwargs)
        self.bind("<MouseWheel>", self._on_mousewheel)
        # For Linux
        self.bind("<Button-4>", self._on_mousewheel)
        self.bind("<Button-5>", self._on_mousewheel)
        # loose focus to parent on Enter key press
        self.bind("<Return>", lambda _e, p=parent: p.focus_set())

    def configure(self, cnf=None, **kwargs):
        customs = [
            "min_value",
            "max_value",
            "include_min_value",
            "include_max_value",
        ]
        for key in customs:
            value = kwargs.pop(key, None)
            if value is not None:
                setattr(self, key, value)
        super().configure(cnf, **kwargs)
        self.update()

    def is_in_bounds(self, value):
        upper_op = operator.le if self.include_max_value else operator.lt
        lower_op = operator.ge if self.include_min_value else operator.gt
        return upper_op(value, self.max_value) and lower_op(value, self.min_value)

    def update(self, value=None):
        if value is None and not self.get():
            logger.debug(f"Update aborted, {self} deliberately empty.")
            return
        value = value if value is not None else self.get()
        try:
            self.var.set(self.format(value))
        except ValueError:
            logger.warning(
                f"Cannot update {self}: {repr(value)} can't be converted to int"
            )

    @staticmethod
    def format(value):
        value = "{:d}".format(int(value))
        return value

    @property
    def allowed_chars(self):
        allowed = "0123456789"
        if self.min_value < 0:
            allowed += "-"
        return allowed

    def _on_mousewheel(self, event):
        if event is not None:
            logger.debug(f"Event caught by {self}._on_mousewheel handler.")
        try:
            current = int(self.var.get())
        except ValueError:
            convertible = False
        else:
            convertible = True
        if str(self["state"]) == "disabled" or not convertible:
            return  # ignore event if widget is disabled or edition unfinished
        delta = event.delta if sys.platform == "darwin" else int(event.delta / 120)
        current = int(self.var.get())
        updated = current + self.scroll_rate * delta
        if self.is_in_bounds(updated):
            self.var.set(self.format(updated))

    def _validate(self, change, after, before, reason, action_code):
        """Enables only values that cen be interpreted as floats."""
        logger.debug(
            f"Input in {self} validation: change={change}, after={after}, "
            f"before={before}, reason={reason}."
        )
        if reason == "focusin":
            self._previous = before
        if action_code and any(c not in self.allowed_chars for c in change):
            return False
        if "-" in change and "-" in before and action_code:
            return False  # do not allow double sign
        if "-" in after and not after.startswith("-"):
            return False  # only allow sign in the beginning
        if not after and reason == "focusout":
            return False  # do not allow no value
        if reason == "focusout":
            try:
                converted = int(after)
            except ValueError:
                return False
            if not self.is_in_bounds(converted):
                return False
            self.var.set(self.format(after))  # format only on valid "focusout"
        return True

    def _on_invalid(self, change, after, before, reason, action_code):
        """Change value to form accepted by float constructor."""
        logger.debug(
            f"Input in {self} invalid: change={change}, after={after}, "
            f"before={before}, reason={reason}."
        )
        if (
            "-" in change
            and not before.startswith("-")
            and action_code  # not deletion
            and self.min_value < 0
        ):
            after = "-" + before
        elif change == "-" and before.startswith("-") and action_code:
            after = before[1:]
        if after == "-" and self.min_value < 0:
            after = after + "0"
        try:
            converted = int(after)
            if not self.is_in_bounds(converted) and reason == "focusout":
                raise ValueError  # treat out-of-bounds value as invalid on "focusout"
        except ValueError:
            # revert if invalid float
            after = self._previous if reason == "focusout" else before
        else:
            # format only on "focusout"
            after = self.format(after) if reason == "focusout" else after
        self.var.set(after)


class NumericEntry(ttk.Entry):
    """Entry that holds a numeric value. Implements validation and changing value
    on mouse wheel event.

    Parameters
    ----------
    scroll_rate : float
        Value to add/substract to/from current value on scroll wheel event.
        Must not be specified if scroll_factor is given.
    scroll_factor : float
        Value by which to multiply/divide current value on scroll wheel event.
        Must not be specified if scroll_rate is given.
    scroll_modifier : callable[float, int]
        Custom function calculating new value after mouse wheel event.
        Must accept current value and standardized scroll delta value as parameters.

    Raises
    ------
    TypeError
        If both, scroll_rate and scroll_factor are specified.
    """

    def __init__(
        self,
        parent,
        scroll_rate=None,
        scroll_factor=None,
        scroll_modifier=None,
        decimal_digits=4,
        rounding=None,  # or "up" or "down"
        keep_trailing_zeros=False,
        min_value=float("-inf"),
        max_value=float("inf"),
        include_min_value=True,
        include_max_value=True,
        **kwargs,
    ):
        self.scroll_factor = scroll_factor
        self.scroll_rate = scroll_rate
        self.scroll_modifier = scroll_modifier
        self.decimal_digits = decimal_digits
        self.rounding = rounding
        self.keep_trailing_zeros = keep_trailing_zeros
        self.min_value = min_value
        self.max_value = max_value
        self.include_min_value = include_min_value
        self.include_max_value = include_max_value
        kwargs["textvariable"] = kwargs.get("textvariable", None) or tk.StringVar()
        kwargs["validate"] = kwargs.get("validate", None) or "all"
        if "validatecommand" not in kwargs:
            validatecommand = (
                parent.register(self._validate),
                "%S",
                "%P",
                "%s",
                "%V",
                "%d",
            )
            kwargs["validatecommand"] = validatecommand
        if "invalidcommand" not in kwargs:
            invalidcommand = (
                parent.register(self._on_invalid),
                "%S",
                "%P",
                "%s",
                "%V",
                "%d",
            )
            kwargs["invalidcommand"] = invalidcommand
        self.var = kwargs["textvariable"]

        self._previous = ""  # used to recover after invalid "select all + paste"

        super().__init__(parent, **kwargs)
        self.bind("<MouseWheel>", self._on_mousewheel)
        # For Linux
        self.bind("<Button-4>", self._on_mousewheel)
        self.bind("<Button-5>", self._on_mousewheel)
        # loose focus to parent on Enter key press
        self.bind("<Return>", lambda _e, p=parent: p.focus_set())

    def is_in_bounds(self, value):
        upper_op = operator.le if self.include_max_value else operator.lt
        lower_op = operator.ge if self.include_min_value else operator.gt
        return upper_op(value, self.max_value) and lower_op(value, self.min_value)

    def configure(self, cnf=None, **kwargs):
        customs = [
            "scroll_rate",
            "scroll_factor",
            "scroll_modifier",
            "decimal_digits",
            "keep_trailing_zeros",
            "min_value",
            "max_value",
            "include_min_value",
            "include_max_value",
        ]
        for key in customs:
            value = kwargs.pop(key, None)
            if value is not None:
                setattr(self, key, value)
        super().configure(cnf, **kwargs)
        self.update()

    def round(self, value):
        factor = 10 ** self.decimal_digits
        if self.rounding == "up":
            return math.ceil(value * factor) / factor
        elif self.rounding == "down":
            return math.floor(value * factor) / factor
        else:
            return round(value, self.decimal_digits)

    def update(self, value=None):
        if value is None and not self.get():
            logger.debug(f"Update aborted, {self} deliberately empty.")
            return
        value = value if value is not None else self.get()
        value = self.round(value) if isinstance(value, float) else value
        try:
            self.var.set(self.format(value))
        except ValueError:
            logger.warning(
                f"Cannot update {self}: {repr(value)} can't be converted to float"
            )

    @property
    def scroll_factor(self):
        return self._scroll_factor

    @scroll_factor.setter
    def scroll_factor(self, value):
        if value is not None and getattr(self, "scroll_rate", None) is not None:
            raise TypeError("Only one, scroll_rate or scroll_factor may be specified.")
        self._scroll_factor = value

    @property
    def scroll_rate(self):
        return self._scroll_rate

    @scroll_rate.setter
    def scroll_rate(self, value):
        if value is not None and getattr(self, "scroll_factor", None) is not None:
            raise TypeError("Only one, scroll_rate or scroll_factor may be specified.")
        self._scroll_rate = value

    @property
    def scroll_modifier(self):
        if self._scroll_modifier is not None:
            return self._scroll_modifier
        elif getattr(self, "scroll_rate") is not None:
            return lambda v, d, r=self.scroll_rate: v + r * d
        elif getattr(self, "scroll_factor") is not None:
            return lambda v, d, f=self.scroll_factor: v * f ** d
        else:
            return lambda v, d: v

    @scroll_modifier.setter
    def scroll_modifier(self, value):
        self._scroll_modifier = value

    def format(self, value):
        formatter = f"{{:.{self.decimal_digits}f}}"
        value = formatter.format(float(value))
        if not self.keep_trailing_zeros:
            value = value.rstrip("0")  # discard insignificant trailing zeros
        if value.endswith("."):
            value += "0"  # but keep at least one decimal digit
        return value

    def _on_mousewheel(self, event):
        if event is not None:
            logger.debug(f"Event caught by {self}._on_mousewheel handler.")
        try:
            _ = float(self.var.get())
        except ValueError:
            convertible = False
        else:
            convertible = True
        if str(self["state"]) == "disabled" or not convertible:
            return  # ignore event if widget is disabled or edition unfinished
        delta = event.delta if sys.platform == "darwin" else int(event.delta / 120)
        current = float(self.var.get())
        updated = self.scroll_modifier(current, delta)
        updated = self.format(updated)
        if self.is_in_bounds(float(updated)):
            self.var.set(updated)

    @property
    def allowed_chars(self):
        allowed = "0123456789.,"
        if self.min_value < 0:
            allowed += "-"
        return allowed

    def _validate(self, change, after, before, reason, action_code):
        """Enables only values that cen be interpreted as floats."""
        logger.debug(
            f"Input in {self} validation: change={change}, after={after}, "
            f"before={before}, reason={reason}."
        )
        if reason == "focusin":
            self._previous = before
        if action_code and any(c not in self.allowed_chars for c in change):
            return False
        if (
            any(c in ".," for c in change)
            and any(c in ".," for c in before)
            and any(c in ".," for c in after)
        ):
            return False  # do not allow double decimal separator
        if "-" in change and "-" in before and action_code:
            return False  # do not allow double sign
        if "-" in after and not after.startswith("-"):
            return False  # only allow sign in the beginning
        if after in ".,-" or after.endswith((".", ",")):
            # includes also unfinished negative float
            return reason != "focusout"  # consider it invalid only when typing is over
        if not after and reason == "focusout":
            return False  # do not allow no value
        if reason == "focusout":
            try:
                converted = float(after)
            except ValueError:
                return False
            if not self.is_in_bounds(converted):
                return False
            self.var.set(self.format(after))  # format only on valid "focusout"
        return True

    def _on_invalid(self, change, after, before, reason, action_code):
        """Change value to form accepted by float constructor."""
        logger.debug(
            f"Input in {self} invalid: change={change}, after={after}, "
            f"before={before}, reason={reason}."
        )
        if (
            "-" in change
            and not before.startswith("-")
            and action_code  # not deletion
            and self.min_value < 0
        ):
            after = "-" + before
        elif change == "-" and before.startswith("-") and action_code:
            after = before[1:]
        if "," in after:
            # consider both, comma and dot, a decimal separator
            after = after.replace(",", ".")
        if after.endswith("."):
            after = after + "0"
        if after == "-" and self.min_value < 0:
            after = after + "0"
        try:
            converted = float(after)
            if not self.is_in_bounds(converted) and reason == "focusout":
                raise ValueError  # treat out-of-bounds value as invalid on "focusout"
        except ValueError:
            # revert if invalid float
            after = self._previous if reason == "focusout" else before
        else:
            # format only on "focusout"
            after = self.format(after) if reason == "focusout" else after
        self.var.set(after)



import pytest
from hypothesis import given, strategies as st

import numpy as np

from tesliper.glassware import spectra as sp


@pytest.mark.parametrize("genre", list(sp.SingleSpectrum._units.keys()))
def test_single_units(genre):
    s = sp.SingleSpectrum(genre, [1, 2], [1, 2])
    assert all(k in s.units for k in ("x", "y", "width", "start", "stop", "step"))


@given(st.integers(min_value=2, max_value=100))
def test_single_length(size):
    np.random.seed(0)  # filler values only, but just to be sure, is shouldn't change
    s = sp.SingleSpectrum("ir", np.random.random(size), np.random.random(size))
    assert len(s) == size



from unittest import mock

import numpy as np
import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

from tesliper.datawork import geometry
from tesliper.datawork.atoms import Atom


@pytest.fixture(scope="module")
def atoms_take():
    return [1, 2, 1, 3]


@pytest.fixture(scope="module")
def atoms_drop():
    return [3, 1, 2, 1]


@pytest.fixture(scope="module")
def one_dim():
    return list(range(4))


@pytest.fixture(scope="module")
def two_dim():
    return [[y * 10 + x for x in range(4)] for y in range(2)]


@pytest.fixture(scope="module")
def three_dim():
    return [
        [[y * 100 + x * 10 + z for z in range(3)] for x in range(4)] for y in range(2)
    ]


@pytest.fixture(scope="module")
def four_dim():
    return [
        [
            [
                [y * 1000 + x * 100 + z * 10, y * 1000 + x * 100 + z * 10 + 1]
                for z in range(3)
            ]
            for x in range(4)
        ]
        for y in range(2)
    ]


@pytest.mark.skip("To be created.")
def test_take_invalid_atom():
    pass


def test_take_absent_atom(atoms_take, two_dim):
    out = geometry.take_atoms(two_dim, atoms_take, 4)
    np.testing.assert_array_equal(out, [])


def test_take_all_atoms(atoms_take, two_dim):
    out = geometry.take_atoms(two_dim, atoms_take, [1, 2, 3])
    np.testing.assert_array_equal(out, two_dim)


@pytest.mark.skip("To be created.")
def test_take_non_matching_sizes():
    pass


def test_take_one_dimension(atoms_take, one_dim):
    out = geometry.take_atoms(one_dim, atoms_take, 1)
    np.testing.assert_array_equal(out, [0, 2])


def test_take_one_dimension_as_enum(atoms_take, one_dim):
    out = geometry.take_atoms(one_dim, atoms_take, Atom.H)
    np.testing.assert_array_equal(out, [0, 2])


def test_take_two_dimensions(atoms_take, two_dim):
    out = geometry.take_atoms(two_dim, atoms_take, 1)
    np.testing.assert_array_equal(out, [[0, 2], [10, 12]])


def test_take_three_dimensions(atoms_take, three_dim):
    out = geometry.take_atoms(three_dim, atoms_take, 1)
    np.testing.assert_array_equal(
        out, [[[0, 1, 2], [20, 21, 22]], [[100, 101, 102], [120, 121, 122]]]
    )


def test_take_four_dimensions(atoms_take, four_dim):
    out = geometry.take_atoms(four_dim, atoms_take, 1)
    np.testing.assert_array_equal(
        out,
        [
            [[[0, 1], [10, 11], [20, 21]], [[200, 201], [210, 211], [220, 221]]],
            [
                [[1000, 1001], [1010, 1011], [1020, 1021]],
                [[1200, 1201], [1210, 1211], [1220, 1221]],
            ],
        ],
    )


@pytest.mark.skip("To be created.")
def test_take_two_dimensions_atoms():
    pass


def test_take_atoms_keeping(atoms_take):
    out = geometry.take_atoms(atoms_take, atoms_take, 1)
    np.testing.assert_array_equal(out, [1, 1])


@pytest.mark.skip("To be created.")
def test_drop_invalid_atom():
    pass


def test_drop_empty_discarded(atoms_drop, two_dim):
    out = geometry.drop_atoms(two_dim, atoms_drop, [])
    np.testing.assert_array_equal(out, two_dim)


def test_drop_absent_atom(atoms_drop, two_dim):
    out = geometry.drop_atoms(two_dim, atoms_drop, 4)
    np.testing.assert_array_equal(out, two_dim)


def test_drop_all_atoms(atoms_drop, two_dim):
    out = geometry.drop_atoms(two_dim, atoms_drop, [1, 2, 3])
    np.testing.assert_array_equal(out, [])


@pytest.mark.skip("To be created.")
def test_non_matching_sizes():
    pass


def test_drop_one_dimension(atoms_drop, one_dim):
    out = geometry.drop_atoms(one_dim, atoms_drop, 1)
    np.testing.assert_array_equal(out, [0, 2])


def test_drop_two_dimensions(atoms_drop, two_dim):
    out = geometry.drop_atoms(two_dim, atoms_drop, 1)
    np.testing.assert_array_equal(out, [[0, 2], [10, 12]])


def test_drop_three_dimensions(atoms_drop, three_dim):
    out = geometry.drop_atoms(three_dim, atoms_drop, 1)
    np.testing.assert_array_equal(
        out, [[[0, 1, 2], [20, 21, 22]], [[100, 101, 102], [120, 121, 122]]]
    )


def test_drop_four_dimensions(atoms_drop, four_dim):
    out = geometry.drop_atoms(four_dim, atoms_drop, 1)
    np.testing.assert_array_equal(
        out,
        [
            [[[0, 1], [10, 11], [20, 21]], [[200, 201], [210, 211], [220, 221]]],
            [
                [[1000, 1001], [1010, 1011], [1020, 1021]],
                [[1200, 1201], [1210, 1211], [1220, 1221]],
            ],
        ],
    )


@pytest.mark.skip("To be created.")
def test_drop_two_dimensions_atoms():
    pass


def test_drop_atoms_dropping(atoms_drop):
    out = geometry.drop_atoms(atoms_drop, atoms_drop, 1)
    np.testing.assert_array_equal(out, [3, 2])


def test_is_triangular_zero():
    assert geometry.is_triangular(0)


def test_is_triangular_one():
    assert geometry.is_triangular(1)


def test_is_triangular_float():
    assert not geometry.is_triangular(0.5)


def test_is_triangular_inf():
    assert not geometry.is_triangular(float("inf"))


def test_is_triangular_negative():
    assert not geometry.is_triangular(-3)


def test_is_triangular_triangular():
    assert geometry.is_triangular(10)


def test_is_triangular_non_triangular():
    assert not geometry.is_triangular(7)


def test_get_triangular_zero():
    assert geometry.get_triangular(0) == 0


def test_get_triangular_one():
    assert geometry.get_triangular(1) == 1


def test_get_triangular_triangular():
    assert geometry.get_triangular(4) == 10


def test_get_triangular_negative():
    with pytest.raises(ValueError):
        geometry.get_triangular(-3)


def test_get_triangular_float():
    with pytest.raises(ValueError):
        geometry.get_triangular(0.5)


def test_get_triangular_inf():
    with pytest.raises(ValueError):
        geometry.get_triangular(float("inf"))


def test_get_triangular_base_zero():
    assert geometry.get_triangular_base(0) == 0


def test_get_triangular_base_one():
    assert geometry.get_triangular_base(1) == 1


def test_get_triangular_base_float():
    with pytest.raises(ValueError):
        geometry.get_triangular_base(0.5)


def test_get_triangular_base_inf():
    with pytest.raises(ValueError):
        geometry.get_triangular_base(float("inf"))


def test_get_triangular_base_negative():
    with pytest.raises(ValueError):
        geometry.get_triangular_base(-3)


def test_get_triangular_base_triangular():
    assert geometry.get_triangular_base(10) == 4


def test_get_triangular_base_non_triangular():
    with pytest.raises(ValueError):
        geometry.get_triangular_base(7)


st_small_floats = st.floats(
    allow_nan=False, allow_infinity=False, max_value=100, min_value=-100
)


@given(
    st.lists(st_small_floats, min_size=9, max_size=9),
    st.lists(st_small_floats, min_size=3, max_size=30),
)
def test_kabsch_rotate(rotation, points):
    assume(not len(points) % 3)
    # random rotation matrix as proposed by:
    # https://math.stackexchange.com/a/1602779/469731
    rotation, _ = np.linalg.qr(np.reshape(rotation, (3, 3)))
    points = np.reshape(points, (-1, 3))
    points = points - points.mean(axis=0)  # zero-centered
    rotated = points @ rotation
    kabsch = geometry.kabsch_rotate(rotated, points)
    np.testing.assert_array_almost_equal(kabsch, points)


@given(
    st.lists(st_small_floats, min_size=18, max_size=18),
    st.lists(st_small_floats, min_size=6, max_size=60),
)
def test_kabsch_rotate_multidim(rotation, points):
    assume(not len(points) % 6)
    rotation = np.array([np.linalg.qr(r)[0] for r in np.reshape(rotation, (2, 3, 3))])
    points = np.reshape(points, (2, -1, 3))
    points = points - np.expand_dims(points.mean(axis=1), 1)  # zero-centered
    rotated = np.array([p @ r for p, r in zip(points, rotation)])
    kabsch = geometry.kabsch_rotate(rotated, points)
    np.testing.assert_array_almost_equal(kabsch, points)


@given(
    st.lists(st_small_floats, min_size=9, max_size=18),
    st.lists(st_small_floats, min_size=3, max_size=30),
)
def test_kabsch_rotate_a_multidim(rotation, points):
    assume(not len(rotation) % 9)
    assume(not len(points) % 3)
    rotation = np.array([np.linalg.qr(r)[0] for r in np.reshape(rotation, (-1, 3, 3))])
    points = np.reshape(points, (-1, 3))
    points = points - points.mean(axis=0)  # zero-centered
    rotated = np.array([points @ r for r in rotation])
    kabsch = geometry.kabsch_rotate(rotated, points)
    np.testing.assert_array_almost_equal(kabsch, [points] * rotation.shape[0])


@given(
    st.lists(st_small_floats, min_size=9, max_size=18),
    st.lists(st_small_floats, min_size=3, max_size=30),
)
def test_kabsch_rotate_b_multidim(rotation, points):
    assume(not len(rotation) % 9)
    assume(not len(points) % 3)
    rotation = np.array([np.linalg.qr(r)[0] for r in np.reshape(rotation, (-1, 3, 3))])
    points = np.reshape(points, (-1, 3))
    points = points - points.mean(axis=0)  # zero-centered
    rotated = np.array([points @ r for r in rotation])
    kabsch = geometry.kabsch_rotate(points, rotated)
    np.testing.assert_array_almost_equal(kabsch, rotated)


@given(st.lists(st_small_floats, min_size=3, max_size=30), st_small_floats)
def test_rmsd(points, error):
    assume(not len(points) % 3)
    points = np.array(points).reshape(-1, 3)
    np.testing.assert_almost_equal(
        geometry.calc_rmsd(points, points + error),
        np.sqrt(3 * error ** 2),  # error adds up from each coordinate
    )


@given(st.integers(min_value=1, max_value=100), st.integers(min_value=1, max_value=100))
def test_fixed_windows(series_size, window_size):
    assume(window_size <= series_size)
    windowed = geometry.fixed_windows(np.empty((series_size,)), window_size)
    assert windowed.shape == (series_size - window_size + 1, window_size)


@given(st.integers(max_value=0))
def test_fixed_windows_wrong_size_value(size):
    with pytest.raises(ValueError, match=r"must be a positive integer"):
        geometry.fixed_windows([], size)


@pytest.mark.parametrize("size", [0.1, "1", tuple(), [], {}, set()])
def test_fixed_windows_wrong_size_type(size):
    with pytest.raises(TypeError, match=r"must be a positive integer"):
        geometry.fixed_windows([], size)


@pytest.mark.parametrize(
    "ens,out",
    [
        ([], []),
        ([1], []),
        (list(range(4)), [[0, 1, 2], [1, 2, 3], [2, 3]]),
        ([1, 2, 10, 20, 21], [[0, 1], [3, 4]]),
    ],
)
def test_sliding_windows(ens, out):
    assert [x.tolist() for x in geometry.stretching_windows(ens, 2)] == out


@pytest.mark.parametrize(
    "ens,out",
    [([], []), ([1], [[0]]), ([1, 2, 10, 20, 21], [[0, 1], [1], [2], [3, 4], [4]])],
)
def test_sliding_windows_keep_hermits(ens, out):
    assert [
        x.tolist() for x in geometry.stretching_windows(ens, 2, keep_hermits=True)
    ] == out


@pytest.mark.parametrize(
    "ens,out",
    [
        ([], []),
        ([1], []),
        (list(range(4)), [[0, 1, 2], [1, 2, 3], [2, 3]]),
        ([1, 2, 3, 20, 21], [[0, 1, 2], [1, 2], [3, 4]]),
    ],
)
def test_sliding_windows_soft_bound(ens, out):
    assert [
        x.tolist() for x in geometry.stretching_windows(ens, 2, hard_bound=False)
    ] == out


@given(size=st.floats(max_value=0, allow_nan=False))
@pytest.mark.parametrize("ens", [[], [1], [1, 1]])
def test_sliding_windows_size_non_positive(ens, size):
    with pytest.raises(ValueError):
        next(geometry.stretching_windows(ens, size))


@pytest.fixture
def rmsd_mock(monkeypatch):
    monkeypatch.setattr(geometry, "center", mock.Mock(side_effect=lambda x: x))
    monkeypatch.setattr(
        geometry, "kabsch_rotate", mock.Mock(side_effect=lambda a, b: a)
    )
    monkeypatch.setattr(
        geometry,
        "calc_rmsd",
        mock.Mock(
            side_effect=lambda a, b: np.abs((a[None, ...] - b).mean(axis=(-2, -1)))
        ),
    )


@pytest.mark.usefixtures("rmsd_mock")
def test_geometry_rmsd_sieve():
    molecule = np.random.normal(1, 3, 15).reshape(5, 3)
    scale = np.array([0.1, 0.2, 3, 3.1, 3.7, 0.1, 5, 0.3, 0.5, 3])
    kept = np.array([1, 0, 1, 0, 1, 1, 1, 0, 1, 1], dtype=bool)
    # as in stretching_windows(np.arange(10), 2, hard_bound=False)
    windows = [np.arange(n, n + 3) for n in range(8)]
    values = molecule[None, ...] + scale[..., None, None]
    np.testing.assert_array_equal(
        geometry.rmsd_sieve(values, windows, threshold=0.5), kept
    )



__author__ = "Michał M. Więcław"
__version_info__ = (0, 9, 2)
__version__ = ".".join(map(str, __version_info__))
# approach coined from suggestions in this thread:
# https://stackoverflow.com/q/458550/11416569

__all__ = ["__author__", "__version_info__", "__version__"]



# IMPORTS
import logging
import os
import tkinter as tk
import tkinter.ttk as ttk
from threading import Thread
from tkinter import messagebox

from .. import tesliper
from .._metadata import __version__
from .components import (
    CheckTree,
    ConformersOverview,
    EnergiesView,
    MaxLevelFilter,
    PopupHandler,
    ReadOnlyText,
    ScrollableFrame,
    ShortExcFormatter,
    SpectraView,
    TextHandler,
    ThreadedMethod,
    WgtStateChanger,
)
from .components.controls import (
    CalculateSpectra,
    ExportData,
    ExtractData,
    FilterEnergies,
    SelectConformers,
)

_DEVELOPMENT = "ENV" in os.environ and os.environ["ENV"] == "prod"


# LOGGER
logger = logging.getLogger(__name__)
home_path = os.path.expanduser("~")
ERROR_LOG_DIR = os.path.join(home_path, "tesliper")
os.makedirs(ERROR_LOG_DIR, exist_ok=True)
error_msg = (
    "Please provide a problem description to Tesliper's "
    'developer along with "tslr_err_log.txt" file, which can be '
    f"found here:\n{ERROR_LOG_DIR}"
)
error_handler = logging.FileHandler(
    os.path.join(ERROR_LOG_DIR, "tslr_err_log.txt"), delay=True
)
error_handler.setLevel(logging.ERROR)
error_handler.setFormatter(
    logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s\n")
)
error_popup_handler = PopupHandler(title_msg="Something unexpected happened! :(")
error_popup_handler.setLevel(logging.ERROR)
error_popup_handler.setFormatter(ShortExcFormatter("%(message)s \n\n" + error_msg))
warning_popup_handler = PopupHandler(title_msg="Sorry!")
warning_popup_handler.setLevel(logging.WARNING)
warning_popup_handler.addFilter(MaxLevelFilter(logging.WARNING))
warning_popup_handler.setFormatter(ShortExcFormatter("%(message)s \n\n"))

handlers = [error_handler, error_popup_handler, warning_popup_handler]

ROOT_LOGGER = logging.getLogger("")
ROOT_LOGGER.setLevel(logging.DEBUG if _DEVELOPMENT else logging.INFO)
for hdlr in handlers:
    ROOT_LOGGER.addHandler(hdlr)
if _DEVELOPMENT:
    # for purposes of debugging
    ROOT_LOGGER.addHandler(tesliper.mainhandler)


# CLASSES
class ViewsNotebook(ttk.Notebook):
    def __init__(self, parent):
        super().__init__(parent)

        self.extract = ConformersOverview(self)
        self.add(self.extract.frame, text="Extracted data")

        self.energies = EnergiesView(self)
        self.add(self.energies.frame, text="Energies list")

        self.spectra = SpectraView(self)
        self.add(self.spectra, text="Spectra view")


class ControlsFrame(ScrollableFrame):
    def __init__(self, parent, extract_view, energies_view, spectra_view, **kwargs):
        super(ControlsFrame, self).__init__(parent, **kwargs)
        tk.Grid.columnconfigure(self, 1, weight=1)

        self.extract = ExtractData(self.content, view=extract_view)
        self.extract.grid(column=0, row=0, sticky="new")
        self.export = ExportData(self.content)
        self.export.grid(column=0, row=1, sticky="new")
        self.select = SelectConformers(self.content, view=extract_view)
        self.select.grid(column=0, row=2, sticky="new")
        self.filter = FilterEnergies(parent=self.content, view=energies_view)
        self.filter.grid(column=0, row=3, sticky="new")
        self.calculate = CalculateSpectra(self.content, view=spectra_view)
        self.calculate.grid(column=0, row=4, sticky="new")


class TesliperApp(tk.Tk):
    def __init__(self):
        super().__init__()
        self.title(f"Tesliper v. {__version__}")
        self.thread = Thread()
        self.changer = WgtStateChanger(self)

        self.protocol("WM_DELETE_WINDOW", self.on_closing)

        tk.Grid.columnconfigure(self, 1, weight=1)
        tk.Grid.rowconfigure(self, 0, weight=1)

        self.tesliper = tesliper.Tesliper()
        self.notebook = ViewsNotebook(self)
        self.notebook.grid(column=1, row=0, sticky="nswe")
        self.controls = ControlsFrame(
            self,
            extract_view=self.notebook.extract,
            energies_view=self.notebook.energies,
            spectra_view=self.notebook.spectra,
        )
        self.controls.grid(column=0, row=0, sticky="nswe")

        # Log & Bar frame
        bottom_frame = tk.Frame(self)
        bottom_frame.grid(column=0, row=1, columnspan=2, sticky="nswe")
        tk.Grid.columnconfigure(bottom_frame, 1, weight=1)
        tk.Grid.rowconfigure(bottom_frame, 0, weight=1)

        # Progress bar
        self.progbar = ttk.Progressbar(
            bottom_frame, length=185, orient=tk.HORIZONTAL, mode="determinate"
        )
        self.progbar.grid(column=0, row=0, sticky="sw")
        self.progtext = tk.StringVar()
        self.progtext.set("Idle.")
        self.proglabel = ttk.Label(
            bottom_frame, textvariable=self.progtext, anchor="w", foreground="gray"
        )
        self.proglabel.grid(column=1, row=0, sticky="swe")

        # Log window
        # displayed in separate, optional window
        self.log = ReadOnlyText(self, width=50, height=34, wrap=tk.WORD)
        ttk.Button(bottom_frame, text="Display log", command=self.log.show).grid(
            column=2, row=0, sticky="se"
        )

        # Logger & handlers
        self.logger = logger
        text_handler = TextHandler(self.log)
        text_handler.setLevel(logging.INFO)
        text_handler.addFilter(MaxLevelFilter(logging.INFO))

        text_warning_handler = TextHandler(self.log)
        text_warning_handler.setLevel(logging.WARNING)
        text_warning_handler.addFilter(MaxLevelFilter(logging.WARNING))
        text_warning_handler.setFormatter(
            logging.Formatter("%(levelname)s: %(message)s")
        )

        text_error_handler = TextHandler(self.log)
        text_error_handler.setLevel(logging.ERROR)
        text_error_handler.setFormatter(
            ShortExcFormatter("ERROR! %(message)s \n" + error_msg)
        )
        text_handlers = [
            text_error_handler,
            text_handler,
            text_warning_handler,
        ]
        for handler in text_handlers:
            ROOT_LOGGER.addHandler(handler)

        self.bind("<<DataExtracted>>", lambda _: self.changer.set_states(), add="+")
        self.bind("<<KeptChanged>>", lambda _: self.changer.set_states(), add="+")

        self.logger.info(
            "Welcome to Tesliper:\n" "Theoretical Spectroscopist Little Helper!"
        )
        try:
            iconpath = os.path.abspath(os.path.realpath(__file__))
            iconpath = os.path.split(os.path.split(iconpath)[0])[0]
            self.iconbitmap(os.path.join(iconpath, "tesliper.ico"))
            self.log.window.iconbitmap(os.path.join(iconpath, "tesliper.ico"))
        except tk.TclError:
            self.logger.warning("Cannot load icon.")

        self.changer.set_states()

    def report_callback_exception(self, exc, val, tb):
        self.logger.critical("An unexpected error occurred.", exc_info=True)

    @ThreadedMethod(progbar_msg="Loading session...")
    def new_tesliper(self, source=None):
        if not source:
            self.tesliper = tesliper.Tesliper()
        else:
            self.tesliper = tesliper.Tesliper.load(source)
            view = self.notebook.extract
            for file, data in self.tesliper.conformers.items():
                view.insert("", tk.END, text=file)
            self.event_generate("<<DataExtracted>>")

    def new_session(self):
        if self.tesliper and self.tesliper.conformers:
            pop = messagebox.askokcancel(
                message="This action will clear the current session "
                "And any unsaved changes will be lost!\n"
                "Would you like to proceed?",
                title="Unsaved changes will be lost!",
                icon="warning",
                default="cancel",
            )
            if not pop:
                return
        self.tesliper.clear()
        self.event_generate("<<Clear>>")
        self.changer.set_states()

    def on_closing(self):
        if self.thread.is_alive():
            quit_ = messagebox.askyesno(
                message="Tesliper is still running an operation. "
                "Do you wish to force exit?",
                title="Exit?",
                icon="warning",
                default="no",
            )
            if not quit_:
                return
        self.destroy()



"""Interface for witing data to disk.

This module contains :func:`.writer` factory function that enables to dynamically create
a writer object that's responsible for saving data in a desired output format.
:func:`.writer` instantiates a subclass of :class:`WriterBase`, an Abstract Base Class
also defined here. :class:`WriterBase` provides an interface for all serial data writers
(objects that export conformers' data to multiple files) used by ``tesliper``.

:class:`WriterBase` expects it's subclasses to provide an *extention* class attribute,
which is used as an extension of files produced by this particular writer, and also as
an identifier for the output format, used by the :func:`.writer` factory function.
``tesliper`` is shipped with four such writers: :class:`.TxtWriter` for writting to .txt
files, :class:`.CsvWriter` for writting in CSV format, :class:`.XlsxWriter` for
creating Excel files, and :class:`.GjfWriter` for preparing Gaussian input files.

You may want to export your data to other file formats - in such case you will need to
implement your own writer. To do this, subclass :class:`WriterBase`, provide it's
*extension* as mentioned above, and implement writing methods for data you intend to
support in your writer. The table below lists these methods, along with a brief
description and :class:`.DataArray`-like object, for which the method will be called by
writer's :meth:`~.WriterBase.write` method.

.. list-table:: Methods used by default to write certain data
    :header-rows: 1

    * - Writer's Method
      - Description
      - Associated array
    * - :meth:`~.WriterBase.generic`
      - Generic data: any genre that provides one value for each conformer.
      - :class:`.DataArray`, :class:`.IntegerArray`,
        :class:`.FloatArray`, :class:`.BooleanArray`, :class:`.InfoArray`.
    * - :meth:`~.WriterBase.overview`
      - General information about conformers: energies, imaginary frequencies,
        stoichiometry.
      - :class:`.Energies`
    * - :meth:`~.WriterBase.energies`
      - Detailed information about conformers' relative energy,
        including calculated populations
      - :class:`.Energies`
    * - :meth:`~.WriterBase.single_spectrum`
      - A spectrum - calculated for single conformer or averaged.
      - :class:`.SingleSpectrum`
    * - :meth:`~.WriterBase.spectral_data`
      - Data related to spectral activity, but not convertible to spectra.
      - :class:`.SpectralData`
    * - :meth:`~.WriterBase.spectral_activities`
      - Data that may be used to simulate conformers' spectra.
      - :class:`.SpectralActivities`
    * - :meth:`~.WriterBase.spectra`
      - Spectra for multiple conformers.
      - :class:`.Spectra`
    * - :meth:`~.WriterBase.transitions`
      - Electronic transitions from ground to excited state, contributing to each band.
      - :class:`.Transitions`
    * - :meth:`~.WriterBase.geometry`
      - Geometry (positions of atoms in space) of conformers.
      - :class:`.Geometry`

.. note::

    These methods are not abstract methods, but will still raise a
    ``NotImplementedError`` if called. This is to let you omit implementation of methods
    you don't need or wouldn't make sense for the particular format and still provide an
    abstract interface. ``tesliper`` takes advantage of this in it's implementation of
    :class:`.GjfWriter`, which only implements :meth:`~.GjfWriter.geometry` method,
    because export of, e.g. a calculated spectrum as a Gaussian input would be
    pointless.

Writer object decides which of these methods to call based on the type of each
:class:`.DataArray`-like object passed to the :meth:`~.WriterBase.write` method. For
some of them, it also passes additional :class:`.DataArray`-like objects, referred to as
*extras*, e.g. correspomding :class:`.Bands` for spectral data. See documentation for
particular method to learn, which of its parameters are mandatory, which are optional,
and which should expect ``None`` as a possible value of *extra*.

When implementing one of these methods in your writer, you should take care of opening
and closing file files, formatting data you export, and writing to the file. For the
first part you may use one of the helper methods that provide a ready-to-use file
handles: :meth:`~.WriterBase._iter_handles` for writing to many files in batch or
:meth:`~.WriterBase._get_handle` for writing to one file only. Both require a template
that will be used to generate filename for produced files. To learn more about how these
templates are handled by ``tesliper``, see :meth:`~.WriterBase.make_name` documentation.

As mentioned before, writer object uses type of the :class:`.DataArray`-like object (or,
more precisely, a name of its class) to decide which method to use for writing to disk.
If you introduce a new subclass of :class:`.DataArray` for handling some genres, you
will need to tell the Writer class, how it should handle these new objects. This is done
by implementing a custom handler method. It's name should begin with an underscore,
followed by the name of your subclass in lower case, followed by "_handler". Also, it
should take two parameters: *data* and *extras*. First one is a list of instances of
your subclass, second one is a dictionary of special-case genres, both retrieved from
arguments given to :meth:`~.WriterBase.write` method (for details on which genres as
treated as special cases, see :meth:`~.WriterBase.distribute_data`). Handler is
responsible for calling appropriate writing method with arguments it needs.

Here is an example: let's assume you have implemented a custom :class:`.DataArray`
subclass for "ldip" and "lrot" genres with some additional functionality, but you'd like
``tesliper`` to treat it as the original :class:`.ElectronicActivities` class for
purposes of writing to disk.

.. code-block:: python

    class LengthActivities(ElectronicActivities):
        associated_genres = ("ldip", "lrot")
        ...  # custom functionality implemented here

    class UpdatedTxtWriter(TxtWriter):
        extension = "txt"

        def _lengthactivities_handler(self, data, extras):
            # written like ``ElectronicActivities``, so just delegate to its handler
            self._electronicactivities_handler(data, extras)

If you'd like to treat this new subclass differently, then you should provide a custom
writting method for this kind of data:

.. code-block:: python

    class UpdatedTxtWriter(TxtWriter):
        extension = "txt"

        def length_activities(
            self,
            band: Bands,
            data: List[LengthActivities],
            name_template: Union[str, Template] = "${conf}.${cat}-${det}.${ext}",
        ):
            # we will use ``_iter_handles`` method for opening/closing files
            template_params = {"genre": band.genre, "cat": "activity", "det": "length"}
            handles = self._iter_handles(band.filenames, name_template, template_params)
            # we will iterate conformer by conformer
            values = zip(*[arr.values for arr in data])
            for values, handle in zip(values, handles):
                ...  # writting logic

        def _lengthactivities_handler(self, data, extras):
            self.length_activities(band=extras["wavelengths"], data=data)

In both cases ``UpdatedTxtWriter`` will be picked by the :func:`.writer` instead of the
original :class:`.TxtWriter`, thanks to the automatic registration done by the base
class :class:`.WriterBase`.

.. warning::

    If ``extension = "txt"`` line would be omitted in the ``UpdatedTxtWriter``
    definition, it would be picked by the :func:`.writer` for "txt" format anyway,
    because ``extension``'s value would be inherited from :class:`.TxtWriter`.
    If you want to prevent this, you can provide a falsy value for the ``extension``
    class attribute, i.e. an empty string or ``None``.
    If your custom writer should still use the same extension as one of the default
    writers, provide ``extension`` also as an instance-level attribute:

    .. code-block:: python

        class UpdatedTxtWriter(TxtWriter):
            extension = ""  # do not register

            def __init__(self, destination, mode):
                super().__init__(destination, mode)
                self.extension = "txt"  # use in generated filenames
"""
import logging as lgg
from abc import ABC, abstractmethod
from contextlib import contextmanager
from pathlib import Path
from string import Formatter, Template
from typing import (
    IO,
    Any,
    AnyStr,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)

from ..glassware.arrays import (
    Bands,
    BooleanArray,
    DataArray,
    ElectronicActivities,
    ElectronicData,
    Energies,
    FloatArray,
    Geometry,
    InfoArray,
    IntegerArray,
    ScatteringActivities,
    ScatteringData,
    SpectralActivities,
    SpectralData,
    Transitions,
    VibrationalActivities,
    VibrationalData,
)
from ..glassware.spectra import SingleSpectrum, Spectra

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


_WRITERS: Dict[str, Type["WriterBase"]] = {}


def writer(
    fmt: str, destination: Union[str, Path], mode: str = "x", **kwargs
) -> "WriterBase":
    """Factory function that returns concrete implementation of :class:`.WriterBase`
    subclass, most recently defined for export to *fmt* file format.

    Parameters
    ----------
    fmt : str
        File format, to which export will be done.
    destination : Union[str, Path]
        Path to file or direcotry, to which export will be done.
    mode : str
        Specifies how writing to file should be handled. Should be one of
        characters: "a" (append to existing file), "x" (only write if file doesn't
        exist yet), or "w" (overwrite file if it already exists). Defaults to "x".
    kwargs
        Any additional keword arguments will be passed as-is to the constructor of the
        retrieved :class:`.WriterBase` subclass.

    Returns
    -------
    WriterBase
        Initialized :class:`.WriterBase` subclass most recently defined for export to
        *fmt* file format.

    Raises
    ------
    ValueError
        If :class:`.WriterBase` subclass for export to *fmt* file format was not
        defined.
    """
    try:
        return _WRITERS[fmt](destination, mode, **kwargs)
    except KeyError:
        message = f"Unknown file format: {fmt}."
        if fmt.startswith(".") and fmt[1:] in _WRITERS:
            message += f" Did you mean '{fmt[1]}'?"
        raise ValueError(message)


_GenericArray = Union[DataArray, IntegerArray, FloatArray, BooleanArray, InfoArray]


# CLASSES
class WriterBase(ABC):
    """Base class for writers that handle export process based on genre of exported
    data.

    Subclasses should provide an :attr:`.extension` class-level attribute and writting
    methods that subclass intend to support (see below). Value of :attr:`.extension`
    will be used to register subclass as a default writer for export to files that this
    value indicates ("txt", "csv", *etc.*). Not providing value for this attribute
    results in a ``TypeError`` exception. If subclass should not be registered, use
    an empty string as the attribute's value.

    :class:`.WriterBase` provides a :meth:`.write` method for writing arbitrary
    :class:`.DataArray`-like objects to disk. It dispatches those objects to appropriate
    writing methods, based on their type. Those writing methods are:

        | :meth:`.generic`,
        | :meth:`.overview`,
        | :meth:`.energies`,
        | :meth:`.single_spectrum`,
        | :meth:`.spectral_data`,
        | :meth:`.spectral_activities`,
        | :meth:`.spectra`,
        | :meth:`.transitions`,
        | :meth:`.geometry`.

    To learn more about implementing custom writers, see their documentation and
    :mod:`.writer_base` documentation or :ref:`extend` section.
    """

    _header = dict(
        zpecorr="Zero-point Corr.",
        tencorr="Thermal Corr.",
        stoichiometry="Stoichiometry",
        entcorr="Enthalpy Corr.",
        last_read_geom="Last Read Geometry",
        optimized_geom="Optimized Geometry",
        input_geom="Input Geometry",
        optimization_completed="Optimized",
        command="Command",
        multiplicity="Multiplicity",
        transitions="Transitions",
        gibcorr="Gibbs Corr.",
        charge="Charge",
        normal_termination="Normal Termination",
        filenames="Filename",
        freq="Frequencies",
        mass="Red. masses",
        frc="Frc consts",
        ramanactiv="Raman Activ",
        depolarp="Depolar (P)",
        depolaru="Depolar (U)",
        ramact="RamAct",
        depp="Dep-P",
        depu="Dep-U",
        alpha2="Alpha2",
        beta2="Beta2",
        alphag="AlphaG",
        gamma2="Gamma2",
        delta2="Delta2",
        cid1="CID1",
        raman2="Raman2",
        roa2="ROA2",
        cid2="CID2",
        raman3="Raman3",
        roa3="ROA3",
        cid3="CID3",
        rc180="RC180",
        rot="Rot. Str.",
        dip="Dip. Str.",
        roa1="ROA1",
        raman1="Raman1",
        ex_en="Excit. Energy",
        wavelen="Wavelength",
        vrot="Rot.(velo)",
        lrot="Rot. (len)",
        vosc="Osc.(velo)",
        losc="Osc. (len)",
        vdip="Dip. (velo)",
        ldip="Dip. (length)",
        iri="IR Int.",
        emang="E-M Angle",
        eemang="E-M Angle",
        zpe="Zero-point",
        ten="Thermal",
        ent="Enthalpy",
        gib="Gibbs",
        scf="SCF",
    )

    _formatters = dict(
        ir="{:> .4e}",
        vcd="{:> .4e}",
        uv="{:> .4e}",
        ecd="{:> .4e}",
        raman="{:> .4e}",
        roa="{:> .4e}",
        zpecorr="{:> 10.4f}",
        tencorr="{:> 10.4f}",
        entcorr="{:> 10.4f}",
        gibcorr="{:> 10.4f}",
        stoichiometry="{}",
        last_read_geom="{}",
        optimized_geom="{}",
        input_geom="{}",
        optimization_completed="{}",
        command="{}",
        multiplicity="{:^ 12d}",
        charge="{:^ 6d}",
        transitions="{}",
        normal_termination="{}",
        filenames="{}",
        rot="{:> 10.4f}",
        dip="{:> 10.4f}",
        roa1="{:> 10.4f}",
        raman1="{:> 10.4f}",
        vrot="{:> 10.4f}",
        lrot="{:> 10.4f}",
        vosc="{:> 10.4f}",
        losc="{:> 10.4f}",
        vdip="{:> 10.4f}",
        ldip="{:> 10.4f}",
        iri="{:> 10.4f}",
        emang="{:> 10.4f}",
        eemang="{:> 10.4f}",
        zpe="{:> 13.4f}",
        ten="{:> 13.4f}",
        ent="{:> 13.4f}",
        gib="{:> 13.4f}",
        scf="{:> 13.4f}",
        ex_en="{:> 13.4f}",
        freq="{:> 10.2f}",
        wavelen="{:> 10.2f}",
        mass="{:> 11.4f}",
        frc="{:> 10.4f}",
        depolarp="{:> 11.4f}",
        depolaru="{:> 11.4f}",
        ramanactiv="{:> 10.4f}",
        ramact="{:> 10.4f}",
        depp="{:> 9.4f}",
        depu="{:> 9.4f}",
        alpha2="{:> 9.4f}",
        beta2="{:> 9.4f}",
        alphag="{:> 9.4f}",
        gamma2="{:> 9.4f}",
        delta2="{:> 9.4f}",
        cid1="{:> 8.3f}",
        raman2="{:> 8.3f}",
        roa2="{:> 8.3f}",
        cid2="{:> 8.3f}",
        raman3="{:> 8.3f}",
        roa3="{:> 8.3f}",
        cid3="{:> 8.3f}",
        rc180="{:> 8.3f}",
    )

    _excel_formats = dict(
        zpecorr="0.0000",
        tencorr="0.0000",
        entcorr="0.0000",
        gibcorr="0.0000",
        multiplicity="{0",
        charge="0",
        stoichiometry="",
        last_read_geom="",
        optimized_geom="",
        input_geom="",
        optimization_completed="",
        command="",
        transitions="",
        normal_termination="",
        filenames="",
        freq="0.0000",
        mass="0.0000",
        frc="0.0000",
        ramanactiv="0.0000",
        depolarp="0.0000",
        depolaru="0.0000",
        ramact="0.0000",
        depp="0.0000",
        depu="0.0000",
        alpha2="0.0000",
        beta2="0.0000",
        alphag="0.0000",
        gamma2="0.0000",
        delta2="0.0000",
        cid1="0.000",
        raman2="0.000",
        roa2="0.000",
        cid2="0.000",
        raman3="0.000",
        roa3="0.000",
        cid3="0.000",
        rc180="0.000",
        rot="0.0000",
        dip="0.0000",
        roa1="0.000",
        raman1="0.000",
        ex_en="0.0000",
        wavelen="0.0000",
        vrot="0.0000",
        lrot="0.0000",
        vosc="0.0000",
        losc="0.0000",
        vdip="0.0000",
        ldip="0.0000",
        iri="0.0000",
        emang="0.0000",
        eemang="0.0000",
        zpe="0.000000",
        ten="0.000000",
        ent="0.000000",
        gib="0.000000",
        scf="0.00000000",
    )

    energies_order = "zpe ten ent gib scf".split(" ")
    """Default order, in which energy-related data is written to files."""

    @property
    @classmethod
    @abstractmethod
    def extension(cls) -> Optional[str]:
        return ""

    extension.__doc__ = """
        Identifier of this writer, indicating the format of files generated,
        and a default extension of those files used by the :meth:`.make_name` method.
        A falsy value, i.e. an empty string or ``None`` prevents this writer from
        being registered and used by :func:`.writer` factory function.

        Returns
        -------
        str
            Default extension of files generated by this writer and it's identifier.
        """

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.extension:
            _WRITERS[cls.extension] = cls

    def __init__(self, destination: Union[str, Path], mode: str = "x"):
        """
        Parameters
        ----------
        destination: str or pathlib.Path
            Directory, to which generated files should be written.
        mode: str
            Specifies how writing to file should be handled. Should be one of
            characters: 'a' (append to existing file), 'x' (only write if file doesn't
            exist yet), or 'w' (overwrite file if it already exists).
        """
        self.mode = mode
        self.destination = destination
        self._handle = None

    @property
    def mode(self):
        """Specifies how writing to file should be handled. Should be one of characters:
        "a", "x", or "w".
        "a" - append to existing file;
        "x" - only write if file doesn't exist yet;
        "w" - overwrite file if it already exists.

        Raises
        ------
        ValueError
            If given anything other than "a", "x", or "w".
        """
        return self._mode

    @mode.setter
    def mode(self, mode):
        if mode not in ("a", "x", "w"):
            raise ValueError("Mode should be 'a', 'x', or 'w'.")
        self._mode = mode

    def check_file(self, file: Union[str, Path]) -> Path:
        file = Path(file)
        if not file.exists() and self.mode == "a":
            raise FileNotFoundError(
                "Mode 'a' was specified, but given file doesn't exist."
            )
        elif file.exists() and self.mode == "x":
            raise FileExistsError(
                "Mode 'x' was specified, but given file already exists."
            )
        elif not file.parent.exists():
            raise FileNotFoundError("Parent directory of specified file doesn't exist.")
        else:
            logger.debug(f"File {file} ok for writing.")
            return file

    @property
    def destination(self) -> Path:
        """pathlib.Path: Directory, to which generated files should be written.

        Raises
        ------
        FileNotFoundError
            If given destination doesn't exist or is not a directory.
        """
        return vars(self)["destination"]

    @destination.setter
    def destination(self, destination: Union[str, Path]) -> None:
        destination = Path(destination)
        if not destination.is_dir():
            raise FileNotFoundError(
                "Given destination doesn't exist or is not a directory."
            )
        vars(self)["destination"] = destination

    @staticmethod
    def distribute_data(data: List) -> Tuple[Dict[str, List], Dict[str, Any]]:
        """Sorts given data by genre category for use by specialized writing methods.

        Returns
        -------
        distr : dict
            Dictionary with :class:`.DataArray`-like objects, sorted by their type.
            Each {key: value} pair is {name of the type in lowercase format:
            list of :class:`.DataArray` objects of this type}.
        extras : dict
            Spacial-case genres: extra information used by some writer methods
            when exporting data. Available {key: value} pairs (if given in *data*) are:

            | corrections: dict of {"energy genre": :class:`.FloatArray`},
            | frequencies: :class:`.Bands`,
            | wavelengths: :class:`.Bands`,
            | excitation: :class:`.Bands`,
            | stoichiometry: :class:`.InfoArray`,
            | charge: :class:`.IntegerArray`,
            | multiplicity: :class:`.IntegerArray`
        """
        distr: Dict[str, List] = dict()
        extras: Dict[str, Any] = dict()
        for obj in data:
            if obj.genre.endswith("corr"):
                corrs = extras["corrections"] = extras.get("corrections", dict())
                corrs[obj.genre[:3]] = obj
            elif obj.genre == "freq":
                extras["frequencies"] = obj
            elif obj.genre == "wavelen":
                extras["wavelengths"] = obj
            elif obj.genre == "ex_en":
                extras["excitation"] = obj
            elif obj.genre == "stoichiometry":
                extras["stoichiometry"] = obj
            elif obj.genre == "charge":
                extras["charge"] = obj
            elif obj.genre == "multiplicity":
                extras["multiplicity"] = obj
            else:
                name = type(obj).__name__.lower()
                values = distr[name] = distr.get(name, list())
                values.append(obj)
        return distr, extras

    def make_name(
        self,
        template: Union[str, Template],
        conf: str = "",
        num: Union[str, int] = "",
        genre: str = "",
        cat: str = "",
        det: str = "",
        ext: str = "",
    ) -> str:
        """Create filename using given template and given or global values
        for known identifiers. The identifier should be used in the template as
        ``"${identifier}"`` where "identifier" is the name of identifier.
        Available names and their meaning are:

        | ``${ext}`` - appropriate file extension
        | ``${conf}`` - name of the conformer
        | ``${num}`` - number of the file according to internal counter
        | ``${genre}`` - genre of exported data
        | ``${cat}`` - category of produced output
        | ``${det}`` - category-specific detail

        The ``${ext}`` identifier is filled with the value of Writers :attr:`.extension`
        attribute if not explicitly given as parameter to this method's call. Values for
        other identifiers should be provided by the caller.

        Parameters
        ----------
        template : str or string.Template
            Template that will be used to generate filenames. It should contain only
            known identifiers, listed above.
        conf : str
            value for ``${conf}`` identifier, defaults to empty string.
        num : str or int
            value for ``${str}`` identifier, defaults to empty string.
        genre : str
            value for ``${genre}`` identifier, defaults to empty string.
        cat : str
            value for ``${cat}`` identifier, defaults to empty string.
        det : str
            value for ``${det}`` identifier, defaults to empty string.
        ext : str
            value for ``${ext}`` identifier, defaults to empty string.

        Raises
        ------
        ValueError
            If given template or string contains any unexpected identifiers.

        Examples
        --------
        Must be first subclassed and instantiated:

        >>> class MyWriter(WriterBase):
        >>>     extension = "foo"
        >>> wrt = MyWriter("/path/to/some/directory/")

        >>> wrt.make_name(template="somefile.${ext}")
        "somefile.foo"
        >>> wrt.make_name(template="${conf}.${ext}")
        ".foo"  # conf is empty string by default
        >>> wrt.make_name(template="${conf}.${ext}", conf="conformer")
        "conformer.foo"
        >>> wrt.make_name(template="Unknown_identifier_${bla}.${ext}")
        Traceback (most recent call last):
        ValueError: Unexpected identifiers given: bla.
        """
        if isinstance(template, str):
            template = Template(template)
        try:
            return template.substitute(
                conf=conf,
                ext=ext or self.extension,
                num=num,
                genre=genre,
                cat=cat,
                det=det,
            )
        except KeyError as error:
            known = {"conf", "ext", "num", "genre", "cat", "det"}
            # second element of each tuple returned is identifier's name
            ids = {parsed[1] for parsed in Formatter().parse(template.template)}
            raise ValueError(f"Unexpected identifiers given: {ids-known}.") from error

    @contextmanager
    def _get_handle(
        self,
        template: Union[str, Template],
        template_params: dict,
        open_params: Optional[dict] = None,
    ) -> Iterator[IO[AnyStr]]:
        """Helper method for creating files. Given additional kwargs will be passed to
        :meth:`Path.open` method. Implemented as context manager for use with ``with``
        statement.

        Parameters
        ----------
        template : str or string.Template
            Template that will be used to generate filenames.
        template_params : dict
            Dictionary of {identifier: value} for `.make_name` method.
        open_params : dict, optional
            Arguments for :meth:`Path.open` used to open file.

        Yields
        ------
        IO
            file handle, will be closed automatically after ``with`` statement exits

        :meta public:
        """
        open_params = open_params or {}  # empty dict by default
        filename = self.make_name(template=template, **template_params)
        file = self.check_file(self.destination.joinpath(filename))
        with file.open(self.mode, **open_params) as handle:
            self._handle = handle
            yield handle

    def _iter_handles(
        self,
        filenames: Iterable[str],
        template: Union[str, Template],
        template_params: dict,
        open_params: Optional[dict] = None,
    ) -> Iterator[IO[AnyStr]]:
        """Helper method for iteration over generated files. Given additional kwargs
        will be passed to :meth:`Path.open` method.

        Parameters
        ----------
        filenames: list of str
            list of source filenames, used as value for `${conf}` placeholder
            in *name_template*
        template_params : dict
            Dictionary of {identifier: value} for `.make_name` method.
        open_params : dict, optional
            arguments for :meth:`Path.open` used to open file.

        Yields
        ------
        TextIO
            file handle, will be closed automatically on next iteration

        :meta public:
        """
        open_params = open_params or {}  # empty dict by default
        for num, fnm in enumerate(filenames):
            template_params.update({"conf": fnm, "num": num})
            filename = self.make_name(template=template, **template_params)
            file = self.check_file(self.destination.joinpath(filename))
            with file.open(self.mode, **open_params) as handle:
                yield handle

    def _dataarray_handler(self, data: List[DataArray], extras: Dict[str, Any]) -> None:
        self.generic(data)

    def _integerarray_handler(
        self, data: List[IntegerArray], extras: Dict[str, Any]
    ) -> None:
        self.generic(data)

    def _floatarray_handler(
        self, data: List[FloatArray], extras: Dict[str, Any]
    ) -> None:
        self.generic(data)

    def _booleanarray_handler(
        self, data: List[BooleanArray], extras: Dict[str, Any]
    ) -> None:
        self.generic(data)

    def _infoarray_handler(self, data: List[InfoArray], extras: Dict[str, Any]) -> None:
        self.generic(data)

    def _energies_handler(self, data: List[Energies], extras: Dict[str, Any]) -> None:
        self.overview(
            data,
            frequencies=extras.get("frequencies"),
            stoichiometry=extras.get("stoichiometry"),
        )
        for en in data:
            self.energies(
                en, corrections=extras.get("corrections", dict()).get(en.genre)
            )

    def _vibrationalactivities_handler(
        self, data: List[VibrationalActivities], extras: Dict[str, Any]
    ) -> None:
        self.spectral_activities(band=extras["frequencies"], data=data)

    def _scatteringactivities_handler(
        self, data: List[ScatteringActivities], extras: Dict[str, Any]
    ) -> None:
        self.spectral_activities(band=extras["frequencies"], data=data)

    def _electronicactivities_handler(
        self, data: List[ElectronicActivities], extras: Dict[str, Any]
    ) -> None:
        self.spectral_activities(band=extras["wavelengths"], data=data)

    def _vibrationaldata_handler(
        self, data: List[VibrationalData], extras: Dict[str, Any]
    ) -> None:
        self.spectral_data(band=extras["frequencies"], data=data)

    def _scatteringdata_handler(
        self, data: List[ScatteringData], extras: Dict[str, Any]
    ) -> None:
        self.spectral_data(band=extras["frequencies"], data=data)

    def _electronicdata_handler(
        self, data: List[ElectronicData], extras: Dict[str, Any]
    ) -> None:
        self.spectral_data(band=extras["wavelengths"], data=data)

    def _transitions_handler(
        self, data: List[Transitions], extras: Dict[str, Any]
    ) -> None:
        if len(data) > 1:
            raise ValueError(
                "Got multiple *Transitions* objects, but can write contents "
                "of only one such object for .write() call."
            )
        self.transitions(transitions=data[0], wavelengths=extras["wavelengths"])

    def _geometry_handler(self, data: List[Geometry], extras: Dict[str, Any]) -> None:
        if len(data) > 1:
            raise ValueError(
                "Got multiple *Geometry* objects, but can write contents "
                "of only one such object for .write() call."
            )
        self.geometry(
            data[0],
            charge=extras.get("charge"),
            multiplicity=extras.get("multiplicity"),
        )

    def _spectra_handler(self, data: List[Spectra], _extras) -> None:
        for spc in data:
            self.spectra(spc)

    def _singlespectrum_handler(self, data: List[SingleSpectrum], _extras) -> None:
        for spc in data:
            self.single_spectrum(spc)

    def write(self, data: List) -> None:
        """Writes :class:`.DataArray`-like objects to disk, decides how to write them
        based on the type of each object. If some types of given objects are not
        supported by this writer, data of this type is ignored and a warning is emitted.

        Parameters
        ----------
        data : List
            :class:`.DataArray`-like objects that should be written to disk.
        """
        distributed, extras = self.distribute_data(data)
        for name, data_ in distributed.items():
            try:
                handler = getattr(self, f"_{name}_handler")
                handler(data_, extras)
            except (NotImplementedError, AttributeError):
                logger.warning(f"{type(self)} does not handle '{name}' type data.")

    def generic(
        self,
        data: List[_GenericArray],
        name_template: Union[str, Template] = "",
    ):
        """Interface for writing generic data: any that provides one value for each
        conformer. Evoked when handling :class:`.DataArray`, :class:`.IntegerArray`,
        :class:`.FloatArray`, :class:`.BooleanArray`, or :class:`.InfoArray`.

        Parameters
        ----------
        data
            List of objects that provide one value for each conformer.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def overview(
        self,
        energies: Sequence[Energies],
        frequencies: Optional[Bands] = None,
        stoichiometry: Optional[InfoArray] = None,
        name_template: Union[str, Template] = "",
    ):
        """Intercafe for generating an overview of known conformers: values of energies,
        number of imaginary frequencies, and stoichiometry for each conformer. Evoked
        when handling :class:`.Energies` objects.

        Parameters
        ----------
        energies
            List of objects representing different energies genres for each conformer.
            Mandatory in custom implementation.
        frequencies
            :class:`.Bands` of "freq" genre, with list of frequencies for each
            conformer. Mandatory in custom implementation. May be ``None`` when method
            evoked by handler.
        stoichiometry
            Stoichiometry of each conformer. Mandatory in custom implementation. May be
            ``None`` when method evoked by handler.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def energies(
        self,
        energies: Energies,
        corrections: Optional[FloatArray] = None,
        name_template: Union[str, Template] = "",
    ):
        """Interface for writing energies values, and optionally their corrections.
        Evoked when handling :class:`.Energies` objects.

        Parameters
        ----------
        energies
            Conformers' energies. Mandatory in custom implementation.
        corrections
            Correction of energies values. Mandatory in custom implementation.
            May be ``None`` when method evoked by handler.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def single_spectrum(
        self, spectrum: SingleSpectrum, name_template: Union[str, Template] = ""
    ):
        """Interface for writing a single spectrum to disk: calculated for one conformer
        or averaged. Evoked when handling :class:`.SingleSpectrum` objects.

        Parameters
        ----------
        spectrum
            Single calculated spectrum. Mandatory in custom implementation.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def spectral_data(
        self,
        band: Bands,
        data: List[SpectralData],
        name_template: Union[str, Template] = "",
    ):
        """Interface for writing multiple objects with spectral data that is not a
        spectral activity (cannot be converted to signal intensity). Evoked when
        handling one of the: :class:`.VibrationalData`, :class:`.ElectronicData`,
        :class:`.ScatteringData` objects.

        Parameters
        ----------
        band
            Band at which transitions occur for each conformer.
            Mandatory in custom implementation.
        data
            List of objects representing different spectral data genres (but not
            spectral activities). Mandatory in custom implementation.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def spectral_activities(
        self,
        band: Bands,
        data: List[SpectralActivities],
        name_template: Union[str, Template] = "",
    ):
        """Interface for writing multiple objects with spectral activities (data that
        may be converted to signal intensity). Evoked when handling one of the:
        :class:`.VibrationalActivities`, :class:`.ElectronicActivities`,
        :class:`.ScatteringActivities` objects.

        Parameters
        ----------
        band
            Band at which transitions occur for each conformer.
            Mandatory in custom implementation.
        data
            List of objects representing different spectral activities genres.
            Mandatory in custom implementation.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def spectra(self, spectra: Spectra, name_template: Union[str, Template] = ""):
        """Interface for writing a set of spectra of one type calculated for many
        conformers. Evoked when handling :class:`.Spectra` objects.

        Parameters
        ----------
        spectra
            Spectra of one type calculated for multiple conformers.
            Mandatory in custom implementation.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def transitions(
        self,
        transitions: Transitions,
        wavelengths: Bands,
        only_highest: bool = True,
        name_template: Union[str, Template] = "",
    ):
        """Interface for writing single object with electronic transitions data.
        Evoked when handling :class:`.Transitions` objects.

        Parameters
        ----------
        transitions
            List of objects representing different spectral data genres (but not
            spectral_activities). Mandatory in custom implementation.
        wavelengths
            Wavelengths at which transitions occur for each conformer.
            Mandatory in custom implementation.
        only_highest
            Boolean flag indicating if all transitions should be written to disk or only
            these transition that contributes the most for each wavelength/
            May be omitted in custom implementation.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")

    def geometry(
        self,
        geometry: Geometry,
        charge: Optional[Union[IntegerArray, Sequence[int], int]] = None,
        multiplicity: Optional[Union[IntegerArray, Sequence[int], int]] = None,
        name_template: Union[str, Template] = "",
    ):
        """Interface for writing single object with geometry of each conformer.
        Evoked when handling :class:`.Geometry` objects.

        Parameters
        ----------
        geometry
            Positions of atoms in each conformer. Mandatory in custom implementation.
        charge
            Value of each structure's charge. Mandatory in custom implementation.
        multiplicity
            Value of each structure's multiplicity. Mandatory in custom implementation.
        name_template
            Template that defines naming scheme for files generated by this method.
            May be omitted in custom implementation.

        Raises
        ------
        NotImplementedError
            Whenever called, this is an interface that should not be used directly.
        """
        raise NotImplementedError(f"Class {type(self)} does not implement this method.")



"""A ``tesliper``'s main data storage."""
import logging as lgg
from collections import Counter, OrderedDict
from collections.abc import ItemsView, KeysView, ValuesView
from contextlib import contextmanager
from inspect import Parameter
from itertools import chain
from reprlib import recursive_repr
from typing import Callable, Iterable, Optional, Sequence, Union

import numpy as np

from tesliper.exceptions import InconsistentDataError, TesliperError

from .. import datawork as dw
from . import arrays as ar
from .array_base import _ARRAY_CONSTRUCTORS

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# TYPE HINTS
AnyArray = Union[
    ar.DataArray,
    ar.Energies,
    ar.FloatArray,
    ar.FilenamesArray,
    ar.InfoArray,
    ar.BooleanArray,
    ar.IntegerArray,
    ar.Bands,
    ar.VibrationalData,
    ar.ScatteringData,
    ar.ElectronicData,
    ar.VibrationalActivities,
    ar.ScatteringActivities,
    ar.ElectronicActivities,
    ar.Transitions,
    ar.Geometry,
]


# CLASSES
class _KeptItemsView(ItemsView):
    def __init__(self, mapping, indices=False):
        super().__init__(mapping)
        self.indices = indices

    def __contains__(self, item):
        key, value = item
        try:
            kept = self._mapping.kept[self._mapping.index_of(key)]
        except KeyError:
            return False
        else:
            if not kept:
                return False
            else:
                v = self._mapping[key]
                return v is value or v == value

    def __iter__(self):
        indices = self.indices
        for idx, (key, kept) in enumerate(zip(self._mapping, self._mapping.kept)):
            if kept:
                value = self._mapping[key]
                yield key, value if not indices else (idx, key, value)

    def __reversed__(self):
        yield from iter(reversed(list(self)))


class _KeptValuesView(ValuesView):
    def __init__(self, mapping, indices=False):
        super().__init__(mapping)
        self.indices = indices

    def __contains__(self, value):
        for key, kept in zip(self._mapping, self._mapping.kept):
            v = self._mapping[key]
            if (v is value or v == value) and kept:
                return True
        return False

    def __iter__(self):
        indices = self.indices
        for idx, (key, kept) in enumerate(zip(self._mapping, self._mapping.kept)):
            if kept:
                value = self._mapping[key]
                yield value if not indices else (idx, value)

    def __reversed__(self):
        yield from iter(reversed(list(self)))


class _KeptKeysView(KeysView):
    def __init__(self, mapping, indices=False):
        super().__init__(mapping)
        self.indices = indices

    def __contains__(self, key):
        try:
            return self._mapping.kept[self._mapping.index_of(key)]
        except KeyError:
            return False

    def __iter__(self):
        indices = self.indices
        for idx, (key, kept) in enumerate(zip(self._mapping, self._mapping.kept)):
            if kept:
                yield key if not indices else (idx, key)

    def __reversed__(self):
        yield from iter(reversed(list(self)))


class Conformers(OrderedDict):
    """Container for data extracted from quantum chemical software output files.

    Data for each file is stored in the underlying OrderedDict, under the key of
    said file's name. Its values are dictionaries with genres name (as key)
    and appropriate data pairs. Beside this, its essential functionality is
    transformation of stored data to corresponding DataArray objects with
    use of :meth:`.arrayed` method. It provides some control over this transformation,
    especially in terms of including/excluding particular conformers' data
    on creation of new DataArray instance. This type of control is here called
    trimming. Trimming can be achieved by use of various *trim* methods defined
    in this class or by direct changes to :attr:`.kept` attribute. See its
    documentation for more information.

    Attributes
    ----------
    primary_genres
        Class attribute. Data genres considered most important, used as default when
        checking for conformers completeness (see :meth:`.trim_incomplete` method).

    Notes
    -----
    Inherits from collections.OrderedDict.
    """

    primary_genres = tuple(
        "dip rot vosc vrot losc lrot raman1 roa1 scf zpe ent ten gib".split()
    )

    def __init__(
        self,
        *args,
        allow_data_inconsistency: bool = False,
        temperature_of_the_system: float = 298.15,
        **kwargs,
    ):
        """
        Parameters
        ----------
        *args
            list of arguments for creation of underlying dictionary
        allow_data_inconsistency : bool, optional
            specifies if data inconsistency should be allowed in created DataArray
            object instances, defaults to False
        temperature_of_the_system : float, optional
            Temperature of the system in Kelvin units, must be zero or higher.
            Defaults to room temperature = 298.15 K.
        **kwargs
            list of arbitrary keyword arguments for creation of underlying
            dictionary
        """
        self.allow_data_inconsistency = allow_data_inconsistency
        self.temperature = temperature_of_the_system
        self.kept = []
        self.filenames = []
        self._indices = {}
        super().__init__(*args, **kwargs)

    @property
    def temperature(self) -> float:
        """Temperature of the system expressed in Kelvin units.

        Value of this parameter is passed to :term:`data array`\\s created with the
        :meth:`.arrayed` method, provided that the target data array class supports a
        parameter named *t* in it's constructor.

        .. versionadded:: 0.9.1

        Raises
        ------
        ValueError
            if set to a value lower than zero.
        """
        return vars(self)["temperature"]

    @temperature.setter
    def temperature(self, value):
        if value <= 0:
            raise ValueError(
                "Temperature of the system must be higher than absolute zero."
            )
        vars(self)["temperature"] = value

    def clear(self):
        """Remove all items from the Conformers instance."""
        self._kept = []
        self.filenames = []
        self._indices = {}
        self.temperature = 298.15
        super().clear()

    @recursive_repr()
    def __repr__(self) -> str:
        if not self:
            return (
                f"{self.__class__.__name__}("
                f"allow_data_inconsistency={self.allow_data_inconsistency})"
            )
        return (
            f"{self.__class__.__name__}({list(self.items())}, "
            f"allow_data_inconsistency={self.allow_data_inconsistency})"
        )

    def __setitem__(self, key, value):
        try:
            value = dict(value)
        except TypeError as error:
            raise TypeError("Can't convert given value to dictionary.") from error
        except ValueError as error:
            raise ValueError("Can't convert given value to dictionary.") from error
        if key in self:
            index = self._indices[key]
        else:
            index = len(self.filenames)
            self.filenames.append(key)
            self.kept.append(True)
        super().__setitem__(key, value)
        self._indices[key] = index

    def __delitem__(self, key):
        index = self._indices[key]
        super().__delitem__(key)
        del self.filenames[index]
        del self.kept[index]
        del self._indices[key]
        for index, key in enumerate(self.keys()):
            self._indices[key] = index

    def popitem(self, last=True):
        """Remove and return a (key, value) pair from the dictionary.

        Pairs are returned in LIFO order if last is true or FIFO order if false.
        """
        idx = -1 if last else 0
        try:
            key = self.filenames[idx]
            return key, self.pop(key)
        except IndexError:
            raise KeyError("Conforemres is empty.")

    def move_to_end(self, key, last=True):
        """Move an existing element to the end (or beginning if last==False).

        Raises KeyError if the element does not exist.
        """
        idx = self.index_of(key)
        new_idx = 0 if not last else len(self.filenames) - 1
        super().move_to_end(key, last)
        self.filenames.insert(new_idx, self.filenames.pop(idx))
        self._kept.insert(new_idx, self._kept.pop(idx))
        self._indices = {k: i for i, k in enumerate(self.filenames)}

    def copy(self):
        "conformers.copy() -> a shallow copy of conformers"
        cp = self.__class__(
            allow_data_inconsistency=self.allow_data_inconsistency,
            temperature_of_the_system=self.temperature,
            **self,
        )
        cp.kept = self.kept
        return cp

    @property
    def kept(self):
        """List of booleans, one for each conformer stored, defining if
        particular conformers data should be included in corresponding DataArray
        instance, created by :meth:`.arrayed` method. It may be changed by use of trim
        methods, by setting its value directly, or by modification of the
        underlying list. For the first option refer to those methods
        documentation, for rest see the Examples section.

        Returns
        -------
        list of bool
            List of booleans, one for each conformer stored, defining if
            particular conformers data should be included in corresponding
            DataArray instance.

        Raises
        ------
        TypeError
            If assigned values is not a sequence.
            If elements of given sequence are not one of types: bool, int, str.
        ValuesError
            If number of given boolean values doesn't match number of contained
            conformers.
        KeyError
            If any of given string values is not in underlying dictionary keys.
        IndexError
            If any of given integer values is not in range
            0 <= i < number of conformers.

        Examples
        --------

        New list of values can be set in a few ways. Firstly, it is the
        most straightforward to just assign a new list of boolean values to
        the :attr:`.kept` attribute. This list should have the same number of elements
        as the number of conformers contained. A ValueError is raised if it
        doesn't.

        >>> c = Conformers(one={}, two={}, tree={})
        >>> c.kept
        [True, True, True]
        >>> c.kept = [False, True, False]
        >>> c.kept
        [False, True, False]
        >>> c.kept = [False, True, False, True]
        Traceback (most recent call last):
        ...
        ValueError: Must provide boolean value for each known conformer.
        4 values provided, 3 excepted.

        Secondly, list of filenames of conformers intended to be kept may be
        given. Only these conformers will be kept. If given filename is not in
        the underlying Conformers' dictionary, KeyError is raised.

        >>> c.kept = ['one']
        >>> c.kept
        [True, False, False]
        >>>  c.kept = ['two', 'other']
        Traceback (most recent call last):
        ...
        KeyError: Unknown conformers: other.

        Thirdly, list of integers representing conformers indices may be given.
        Only conformers with specified indices will be kept. If one of given integers
        can't be translated to conformer's index, IndexError is raised. Indexing with
        negative values is not supported currently.

        >>> c.kept = [1, 2]
        >>> c.kept
        [False, True, True]
        >>> c.kept = [2, 3]
        Traceback (most recent call last):
        ...
        IndexError: Indexes out of bounds: 3.

        Fourthly, assigning ``True`` or ``False`` to this attribute will mark all
        conformers as kept or not kept respectively.

        >>> c.kept = False
        >>> c.kept
        [False, False, False]
        >>> c.kept = True
        >>> c.kept
        [True, True, True]

        Lastly, list of kept values may be modified by setting its elements
        to True or False. It is advised against, however, as mistake such as
        ``c.kept[:2] = [True, False, False]`` will break some functionality by
        forcibly changing size of :attr:`.kept` list.

        Notes
        -----
        Type of the first element of given sequence is used for dynamic
        dispatch.
        """
        # TODO: Consider making return value immutable.
        return self._kept

    @kept.setter
    def kept(self, blade: Union[Sequence[bool], Sequence[str], Sequence[int], bool]):
        if blade is True or blade is False:
            self._kept = [blade for _ in self.keys()]
            return
        try:
            first = blade[0]
        except (TypeError, KeyError):
            raise TypeError(f"Excepted sequence or boolean, got: {type(blade)}.")
        except IndexError:
            self._kept = [False for _ in self.keys()]
            return  # empty sequence is understood as "keep nothing"
        if isinstance(first, (str, np.str_)):
            blade = set(blade)
            if not blade.issubset(self.keys()):
                raise KeyError(f"Unknown conformers: {', '.join(blade-self.keys())}")
            else:
                self._kept = [fnm in blade for fnm in self.keys()]
        elif isinstance(first, (bool, np.bool_)):
            if not len(blade) == len(self):
                raise ValueError(
                    f"Must provide boolean value for each known conformer. "
                    f"{len(blade)} values provided, {len(self)} excepted."
                )
            else:
                self._kept = [bool(b) for b in blade]  # convert from np.bool_
        elif isinstance(first, (int, np.integer)):
            length = len(self)
            out_of_bounds = [b for b in blade if not 0 <= b < length]
            if out_of_bounds:
                raise IndexError(
                    f"Indexes out of bounds: "
                    f"{', '.join(str(n) for n in out_of_bounds)}."
                )
            else:
                blade = set(blade)
                self._kept = [num in blade for num in range(len(self))]
        else:
            raise TypeError(
                f"Expected sequence of strings, integers or booleans, got: "
                f"{type(first)} as first sequence's element."
            )

    def update(self, other=None, **kwargs):
        """Works like ``dict.update``, but if key is already present, it updates
        dictionary associated with given key rather than assigning new value.
        Keys of dictionary passed as positional parameter (or additional keyword
        arguments given) should be conformers' identifiers and its values should be
        dictionaries of {"genre": values} for those conformers.

        Please note, that values of status genres like 'optimization_completed'
        and 'normal_termination' will be updated as well for such key,
        if are present in given new values.
        """
        if other is not None:
            other = dict(other)
        else:
            other = dict()
        items = chain(other.items(), kwargs.items())
        for key, value in items:
            if key in self:
                self[key].update(value)
            else:
                self[key] = value

    def arrayed(
        self, genre: str, full: bool = False, strict: bool = True, **kwargs
    ) -> AnyArray:
        """Lists requested data and returns as appropriate :class:`.DataArray` instance.

        .. versionadded:: 0.9.1
            The *strict* parameter.

        Parameters
        ----------
        genre
            String representing data genre. Must be one of known genres.
        full
            Boolean indicating if full set of data should be taken, ignoring
            any trimming conducted earlier. Defaults to ``False``.
        strict
            Boolean indicating if additional kwargs that doesn't match signature of data
            array's constructor should cause an exception as normally (``strict =
            True``) or be silently ignored (``strict = False``). Defaults to ``True``.
        kwargs
            Additional keyword parameters passed to data array constructor.
            Any explicitly given parameters will take precedence over automatically
            retrieved and default values.

        Returns
        -------
        DataArray
            Arrayed data of desired genre as appropriate :class:`.DataArray` object.

        Notes
        -----
        For now, the special "filenames" genre always ignores *kwargs*.
        """
        try:
            cls = _ARRAY_CONSTRUCTORS[genre]  # ArrayBase subclasses
        except KeyError:
            raise ValueError(f"Unknown genre '{genre}'.")
        if genre == "filenames":
            # return early if filenames requested
            return cls(
                genre=genre,
                filenames=list(self.kept_keys() if not full else self.keys()),
                allow_data_inconsistency=self.allow_data_inconsistency,
            )
        view = self.kept_items() if not full else self.items()
        array = ((fname, conf, conf[genre]) for fname, conf in view if genre in conf)
        try:
            filenames, confs, values = zip(*array)
        except ValueError:  # if no elements in `array`
            logger.debug(
                f"Array of gerne {genre} requested, but no such data available "
                f"or conformers providing this data were trimmed off. "
                f"Returning an empty array."
            )
            filenames, confs, values = [], [], []
        default_params = cls.get_init_params()
        default_params["genre"] = genre
        default_params["filenames"] = filenames
        default_params["values"] = values
        default_params["allow_data_inconsistency"] = self.allow_data_inconsistency
        init_params = {}
        for key, value in default_params.items():
            if key in kwargs:
                # explicitly given keyword parameters take precedence
                init_params[key] = kwargs.pop(key)
                continue
            if key == "t":
                # if not given explicitly, temperature is taken form self
                init_params[key] = self.temperature
                continue
            if not isinstance(default_params[key], Parameter):
                # if value for parameter is already established, just take it
                init_params[key] = value
                continue
            param_genre = (  # maybe ``key`` is not a param's genre name
                value.genre_getter(genre) if hasattr(value, "genre_getter") else key
            )
            try:
                init_params[key] = [conf[param_genre] for conf in confs]
            except KeyError:
                # can't retrieve ``param_genre`` data from each included conformer
                # set param to its default value or raise an error if it don't have one
                if value.default is not value.empty:
                    init_params[key] = value.default
                else:
                    raise TesliperError(
                        f"One or more conformers does not provide value for "
                        f"'{param_genre}' genre, needed to instantiate {cls.__name__} "
                        "object. You may provide missing values as a keyword parameters"
                        " to the `Conformers.arrayed()` method call."
                    )
            if not init_params[key] and value.default is not value.empty:
                # genre produces an empty array, but parameter has default value
                init_params[key] = value.default
        if kwargs and strict:
            # any kwargs not popped till now are not expected by the ``cls.__init__()``
            # if ``strict`` handling requested, add them anyway to cause an exception
            init_params.update(**kwargs)
        return cls(**init_params)

    def by_index(self, index: int) -> dict:
        """Returns data for conformer on desired index."""
        return self[self.filenames[index]]

    def key_of(self, index: int) -> str:
        """Returns name of conformer associated with given index."""
        return self.filenames[index]

    def index_of(self, key: str) -> int:
        """Return index of given key."""
        try:
            return self._indices[key]
        except KeyError as error:
            raise KeyError(f"No such conformer: {key}.") from error

    def has_genre(self, genre: str, ignore_trimming: bool = False) -> bool:
        """Checks if any of stored conformers contains data of given genre.

        Parameters
        ----------
        genre : str
            Name of genre to test.
        ignore_trimming : bool
            If all known conformers should be considered (``ignore_trimming = True``)
            or only kept ones (``ignore_trimming = False``, default).

        Returns
        -------
        bool
            Boolean value indicating if any of stored conformers contains data
            of genre in question."""
        conformers = self.values() if ignore_trimming else self.kept_values()
        for conformer in conformers:
            if genre in conformer:
                return True
        return False

    def has_any_genre(
        self, genres: Iterable[str], ignore_trimming: bool = False
    ) -> bool:
        """Checks if any of stored conformers contains data of any of given
        genres.

        Parameters
        ----------
        genres : iterable of str
            List of names of genres to test.
        ignore_trimming : bool
            If all known conformers should be considered (``ignore_trimming = True``)
            or only kept ones (``ignore_trimming = False``, default).

        Returns
        -------
        bool
            Boolean value indicating if any of stored conformers contains data
            of any of genres in question."""
        conformers = self.values() if ignore_trimming else self.kept_values()
        for conformer in conformers:
            for genre in genres:
                if genre in conformer:
                    return True
        return False

    def all_have_genres(
        self, genres: Iterable[str], ignore_trimming: bool = False
    ) -> bool:
        """Checks if all stored conformers contains data of given genres.

        Parameters
        ----------
        genres : iterable of str
            List of names of genres to test.
        ignore_trimming : bool
            If all known conformers should be considered (``ignore_trimming = True``)
            or only kept ones (``ignore_trimming = False``, default).

        Returns
        -------
        bool
            Boolean value indicating if each stored conformers contains data
            of all genres in question."""
        genres = set(genres)
        conformers = self.values() if ignore_trimming else self.kept_values()
        for conformer in conformers:
            if genres - conformer.keys():
                return False
        return True

    def trim_incomplete(
        self, wanted: Optional[Iterable[str]] = None, strict: bool = False
    ) -> None:
        """Mark incomplete conformers as "not kept".

        Conformers that does not contain one or more data genres specified as *wanted*
        will be marked as "not kept". If *wanted* parameter is not given, it evaluates
        to :attr:`.primary_genres`. If no conformer contains all *wanted* genres,
        conformers that match the specification most closely are kept. The "closeness"
        is defined by number of conformer's genres matching *wanted* genres in the first
        place (the more, the better) and the position of particular genre in *wanted*
        list in the second place (the closer to the beginning, the better). This
        "match closest" behaviour may be turned off by setting parameter
        *strict* to ``True``. In such case, only conformers containing all *wanted*
        genres will be kept.

        Parameters
        ----------
        wanted
            List of data genres used as completeness reference.
            If not given, evaluates to :attr:`.primary_genres`.
        strict
            Indicates if all *wanted* genres must be present in the kept conformers
            (``strict=True``) or if "match closest" mechanism should be used
            as a fallback (``strict=False``, this is the default).

        Notes
        -----
        Conformers previously marked as "not kept" will not be affected.
        """
        wanted = wanted if wanted is not None else self.primary_genres
        if not strict:
            count = [tuple(g in conf for g in wanted) for conf in self.values()]
            if not count:
                return
            best_match = max(count)
            complete = (match == best_match for match in count)
        else:
            complete = (all(g in conf for g in wanted) for conf in self.values())
        blade = [kept and cmpl for kept, cmpl in zip(self.kept, complete)]
        self._kept = blade

    def trim_imaginary_frequencies(self) -> None:
        """Mark all conformers with imaginary frequencies as "not kept".

        Notes
        -----
        Conformers previously marked as "not kept" will not be affected.
        Conformers that doesn't contain "freq" genre will be treated as not having
        imaginary frequencies.
        """
        dummy = [1]
        for index, conf in enumerate(self.values()):
            freq = np.array(conf.get("freq", dummy))
            if (freq < 0).any():
                self._kept[index] = False

    def trim_non_matching_stoichiometry(self, wanted: Optional[str] = None) -> None:
        """Mark all conformers with stoichiometry other than *wanted* as "not kept".
        If not given, *wanted* evaluates to the most common stoichiometry.

        Parameters
        ----------
        wanted
            Only conformers with same stoichiometry will be kept. Evaluates to the most
            common stoichiometry if not given.

        Notes
        -----
        Conformers previously marked as "not kept" will not be affected.
        Conformers that doesn't contain stoichiometry data are always treated
        as non-matching.
        """
        if not wanted:
            counter = Counter(
                conf["stoichiometry"]
                for conf in self.values()
                if "stoichiometry" in conf
            )
            counts = counter.most_common()
            wanted = counts[0][0] if counts else ""  # no conformer has "stoichiometry"
        for index, conf in enumerate(self.values()):
            if "stoichiometry" not in conf or not conf["stoichiometry"] == wanted:
                self._kept[index] = False

    def trim_not_optimized(self) -> None:
        """Mark all conformers that failed structure optimization as "not kept".

        Notes
        -----
        Conformers previously marked as "not kept" will not be affected.
        Conformers that doesn't contain optimization data are always treated as
        optimized.
        """
        for index, conf in enumerate(self.values()):
            if not conf.get("optimization_completed", True):
                self._kept[index] = False

    def trim_non_normal_termination(self) -> None:
        """Mark all conformers, which calculation job did not terminate normally,
         as "not kept".

        Notes
        -----
        Conformers previously marked as "not kept" will not be affected.
        Conformers that doesn't contain data regarding their calculation job's
        termination are always treated as terminated abnormally.
        """
        for index, conf in enumerate(self.values()):
            if not conf.get("normal_termination", False):
                self._kept[index] = False

    def trim_inconsistent_sizes(self) -> None:
        """Mark as "not kept" all conformers that contain any iterable data genre,
        that is of different length, than in case of majority of conformers.

        Examples
        --------
        >>> c = Conformers(
        ...     one={'a': [1, 2, 3]},
        ...     two={'a': [1, 2, 3]},
        ...     three={'a': [1, 2, 3, 4]}
        ... )
        >>> c.kept
        [True, True, True]
        >>> c.trim_inconsistent_sizes()
        >>> c.kept
        [True, True, False]

        Notes
        -----
        Conformers previously marked as "not kept" will not be affected.
        """
        sizes = {}
        for fname, conf in self.items():
            for genre, value in conf.items():
                if isinstance(value, (np.ndarray, list, tuple)):
                    sizes.setdefault(genre, {})[fname] = len(value)
        maxes = {
            genre: Counter(v for v in values.values()).most_common()[0][0]
            for genre, values in sizes.items()
        }
        for index, fname in enumerate(self.keys()):
            for genre, most_common in maxes.items():
                confs = sizes[genre]
                if fname in confs and not confs[fname] == most_common:
                    self._kept[index] = False

    def trim_to_range(
        self,
        genre: str,
        minimum: Union[int, float] = float("-inf"),
        maximum: Union[int, float] = float("inf"),
        attribute: str = "values",
    ) -> None:
        """Marks as "not kept" all conformers, which numeric value of data
        of specified genre is outside of the range specified by *minimum*
        and *maximum* values.

        Parameters
        ----------
        genre
            Name of genre that should be compared to specified
            minimum and maximum values.
        minimum
            Minimal accepted value - every conformer, which genre value evaluates
            to less than *minimum* will be marked as "not kept".
            Defaults to ``float(-inf)``.
        maximum
            Maximal accepted value - every conformer, which genre value evaluates
            to more than *maximum* will be marked as "not kept".
            Defaults to ``float(inf)``.
        attribute
            Attribute of DataArray of specified *genre* that contains one-dimensional
            array of numeric values. defaults to `"values"`.

        Raises
        ------
        AttributeError
            If DataArray associated with *genre* genre has no attribute *attribute*.
        ValueError
            If data retrieved from specified genre's attribute is not in the form of
            one-dimensional array.
        TypeError
            If comparision cannot be made between elements of specified genre's
            attribute and *minimum* or *maximum* values.

        Notes
        -----
        Conformers previously marked as "not kept" will not be affected.
        """
        try:
            arr = self.arrayed(genre)
            atr = getattr(arr, attribute)
        except AttributeError as error:
            raise AttributeError(
                f"Invalid genre/attribute combination: {genre}/{attribute}. "
                f"Resulting DataArray object has no attribute {attribute}."
            ) from error
        values = np.asarray(atr)
        if values.ndim != 1:
            raise ValueError(
                f"Invalid genre/attribute combination: {genre}/{attribute}. "
                f"DataArray's attribute must contain one-dimensional array of values."
            )
        try:
            in_range = (minimum <= values) & (values <= maximum)
        except TypeError as error:
            raise TypeError(
                f"Cannot compare {type(minimum)} with {type(values[0])}."
            ) from error
        self.kept = arr.filenames[in_range]

    def trim_rmsd(
        self,
        threshold: Union[int, float],
        window_size: Optional[Union[int, float]],
        geometry_genre: str = "last_read_geom",
        energy_genre: str = "scf",
        ignore_hydrogen: bool = True,
        moving_window_strategy: Callable = dw.stretching_windows,
    ) -> None:
        """Marks as "not kept" all conformers that are identical with some other
        conformer, judging by a provided RMSD threshold.

        To minimize computation cost, conformers are compared inside windows, that is a
        subsets of the original list of conformers. Those windows are generated by the
        *moving_window_strategy* function. The recommended strategy, and a default
        value, is :func:`.streaching_windows`, but other are also available:
        :func:`.fixed_windows` and :func:`.pyramid_windows`. This function will be
        called with list of energies for conformers compared and (if it is not ``None``)
        *window_size* parameter.

        With default *moving_window_strategy* conformers, which energy difference (dE)
        is higher than given *window_size* are always treated as different, while those
        with dE smaller than *window_size* and RMSD value smaller than given *threshold*
        are considered identical. From two identical conformers, the one with lower
        energy is "kept", and the other is discarded (marked as "not kept").

        Notes
        -----
        RMSD threshold and size of the energy window should be chosen depending on the
        parameters of conformers' set: number of conformers, size of the conformer,
        its lability, etc. However, *threshold* of 0.5 angstrom and *window_size*
        of 5 to 10 kcal/mol is a good place to start if in doubt.

        Parameters
        ----------
        threshold : int or float
            Maximum RMSD value to consider conformers identical.
        window_size : int or float
            Size of the energy window, in kcal/mol, inside which RMSD matrix is
            calculated. Essentially, a difference in conformers' energy, after which
            conformers are always considered different.
        geometry_genre : str
            Genre of geometry used to calculate RMSD matrix. "last_read_geom" is
            default.
        energy_genre : str
            Genre of energy used to sort and group conformers into windows of given
            energy size. "scf" is used by default.
        ignore_hydrogen : bool
            If hydrogen atom should be discarded before RMSD calculation.
            Defaults to ``True``.
        moving_window_strategy : callable
            Function that generates windows, inside which RMSD comparisions is
            performed.

        Raises
        ------
        InconsistentDataError
            If requested genres does not provide the same set of conformers.
        ValueError
            When called with ``ignore_hydrogen=True`` but requested
            :attr:`.Geometry.atoms` cannot be collapsed to 1-D array.
        """
        energy = self.arrayed(energy_genre)
        geometry = self.arrayed(geometry_genre)
        if not energy.filenames.size == geometry.filenames.size:
            raise InconsistentDataError(
                "Unequal number of conformers in requested geometry and energy genres. "
                "Trim incomplete entries before trimming with :meth:`.trim_rmds`."
            )
        elif not np.array_equal(energy.filenames, geometry.filenames):
            raise InconsistentDataError(
                "Different conformers in requested geometry and energy genres. "
                "Trim incomplete entries before trimming with :meth:`.trim_rmds`."
            )
        if not geometry:
            return  # next steps assume there are some conformers
        if ignore_hydrogen and geometry.atoms.shape[0] > 1:
            # TODO: remove when dw.geometry.select_atoms supplemented
            raise ValueError(
                "Cannot ignore hydrogen atoms if requested conformers do not have "
                "the same order of atoms. This functionality is not supported yet."
            )
        geom = (
            dw.drop_atoms(geometry.values, geometry.atoms[0], dw.atoms.Atom.H)
            if ignore_hydrogen
            else geometry.values
        )
        if window_size is None:
            windows = moving_window_strategy(energy.as_kcal_per_mol)
        else:
            windows = moving_window_strategy(energy.as_kcal_per_mol, window_size)
        wanted = dw.rmsd_sieve(geom, windows, threshold)
        self.kept = geometry.filenames[wanted]

    def select_all(self) -> None:
        """Marks all conformers as 'kept'. Equivalent to ``conformers.kept = True``."""
        self._kept = [True for _ in self._kept]

    def reject_all(self) -> None:
        """Marks all conformers as 'not kept'. Equivalent to
        ``conformers.kept = False``.
        """
        self._kept = [False for _ in self._kept]

    def kept_keys(self, indices: bool = False) -> _KeptKeysView:
        """Equivalent of ``dict.keys()`` but gives view only on conformers marked
        as "kept". Returned view may also provide information on conformers index
        in its Conformers instance if requested with ``indices=True``.

        >>> c = Conformers(c1={"g": 0.1}, c2={"g": 0.2}, c3={"g": 0.3}}
        >>> c.kept = [True, False, True]
        >>> list(c.kept_keys())
        ["c1", "c3"]
        >>> list(c.kept_keys(indices=True))
        [(0, "c1"}), (2, "c3")]

        Parameters
        ----------
        indices : bool
            If resulting Conformers view should also provide index of each conformer.
            Defaults to False.

        Returns
        -------
        _KeptKeysView
            View of kept conformers.
        """
        return _KeptKeysView(self, indices=indices)

    def kept_values(self, indices: bool = False) -> _KeptValuesView:
        """Equivalent of ``dict.values()`` but gives view only on conformers marked
        as "kept". Returned view may also provide information on conformers index
        in its Conformers instance if requested with ``indices=True``.

        >>> c = Conformers(c1={"g": 0.1}, c2={"g": 0.2}, c3={"g": 0.3}}
        >>> c.kept = [True, False, True]
        >>> list(c.kept_values())
        [{"g": 0.1}, {"g": 0.3}]
        >>> list(c.kept_values(indices=True))
        [(0, {"g": 0.1}), (2,  {"g": 0.3})]

        Parameters
        ----------
        indices : bool
            If resulting Conformers view should also provide index of each conformer.
            Defaults to False.

        Returns
        -------
        _KeptValuesView
            View of kept conformers.
        """
        return _KeptValuesView(self, indices=indices)

    def kept_items(self, indices: bool = False) -> _KeptItemsView:
        """Equivalent of ``dict.items()`` but gives view only on conformers marked
        as "kept". Returned view may also provide information on conformers index
        in its Conformers instance if requested with ``indices=True``.

        >>> c = Conformers(c1={"g": 0.1}, c2={"g": 0.2}, c3={"g": 0.3}}
        >>> c.kept = [True, False, True]
        >>> list(c.kept_items())
        [("c1", {"g": 0.1}), ("c3", {"g": 0.3})]
        >>> list(c.kept_items(indices=True))
        [(0, "c1", {"g": 0.1}), (2, "c3", {"g": 0.3})]

        Parameters
        ----------
        indices : bool
            If resulting Conformers view should also provide index of each conformer.
            Defaults to False.

        Returns
        -------
        _KeptItemsView
            View of kept conformers.
        """
        return _KeptItemsView(self, indices=indices)

    @property
    @contextmanager
    def untrimmed(self) -> "Conformers":
        """Temporally remove trimming. Implemented as context manager to use with
        python's 'with' keyword.

        Examples
        --------
        >>> c = Conformers(one={}, two={}, tree={})
        >>> c.kept = [False, True, False]
        >>> with c.untrimmed:
        >>>     c.kept
        [True, True, True]
        >>> c.kept
        [False, True, False]
        """
        blade = self._kept
        self.kept = True
        yield self
        self._kept = blade

    @contextmanager
    def trimmed_to(
        self, blade: Union[Sequence[bool], Sequence[str], Sequence[int], bool]
    ) -> "Conformers":
        """Temporally set trimming blade to given one. Implemented as context manager
        to use with python's 'with' keyword.

        Parameters
        ----------
        blade : bool or sequence of bool, str, or int
            Temporary trimming blade. To better understand how blade setting works,
            see Conformers.kept documentation.

        Examples
        --------
        >>> c = Conformers(one={}, two={}, tree={})
        >>> c.kept = [True, True, False]
        >>> with c.trimmed_to([1, 2]):
        >>>     c.kept
        [False, True, True]
        >>> c.kept
        [True, True, False]
        """
        old_blade = self._kept
        self.kept = blade
        yield self
        self._kept = old_blade

    @property
    @contextmanager
    def inconsistency_allowed(self) -> "Conformers":
        """Temporally sets Conformers' 'allow_data_inconsistency' attribute
        to true. Implemented as context manager to use with python's 'with' keyword.

        Examples
        --------
        >>> c = Conformers(...)
        >>> with c.inconsistency_allowed:
        >>>     # do stuff here while c.allow_data_inconsistency is True
        >>>     c.allow_data_inconsistency
        True
        >>> c.allow_data_inconsistency
        False
        """
        inconsistency = self.allow_data_inconsistency
        self.allow_data_inconsistency = True
        yield self
        self.allow_data_inconsistency = inconsistency



# IMPORTS
import logging as lgg
import queue
import tkinter as tk
import tkinter.ttk as ttk
from collections import defaultdict
from copy import copy
from threading import Thread
from tkinter import messagebox
from tkinter.scrolledtext import ScrolledText
from typing import Iterable, List

# LOGGER
logger = lgg.getLogger(__name__)


# FUNCTIONS
def join_with_and(words: List[str]) -> str:
    """Joins list of strings with "and" between the last two."""
    if len(words) > 2:
        return ", ".join(words[:-1]) + ", and " + words[-1]
    elif len(words) == 2:
        return " and ".join(words)
    elif len(words) == 1:
        return words[0]
    else:
        return ""


# CLASSES
class TextHandler(lgg.Handler):
    def __init__(self, widget, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = widget

    def emit(self, record):
        msg = self.format(record)
        self.widget.insert("end", msg + "\n", record.levelname)
        self.widget.yview("end")


class MaxLevelFilter:
    def __init__(self, max_level):
        self.max_level = max_level

    def filter(self, record):
        return record.levelno <= self.max_level


class ShortExcFormatter(lgg.Formatter):
    def format(self, record):
        record = copy(record)
        record.exc_text = ""
        return super().format(record)

    def formatException(self, ei):
        output = ""
        return output


class PopupHandler(lgg.Handler):
    def __init__(self, *args, title_msg="", **kwargs):
        super().__init__(*args, **kwargs)
        self.title_msg = title_msg

    def emit(self, record):
        msg = self.format(record)
        if record.levelno == lgg.INFO:
            messagebox.showinfo(self.title_msg, msg)
        elif record.levelno == lgg.WARNING:
            messagebox.showwarning(self.title_msg, msg)
        elif record.levelno >= lgg.ERROR:
            messagebox.showerror(self.title_msg, msg)


class ReadOnlyText(ScrolledText):
    def __init__(self, master, **kwargs):
        self.window = tk.Toplevel(master)
        self.hide()
        self.window.title("Tesliper Log")
        self.window.protocol("WM_DELETE_WINDOW", self.hide)
        kwargs["state"] = "disabled"
        super().__init__(self.window, **kwargs)
        self.pack(fill=tk.BOTH, expand=tk.YES)
        self.tag_config("DEBUG", foreground="gray")
        self.tag_config("INFO", foreground="black")
        self.tag_config("WARNING", foreground="dark violet", font="Courier 10 italic")
        self.tag_config("ERROR", foreground="red3")
        self.tag_config("CRITICAL", foreground="red3", font="Courier 10 bold")

    def insert(self, *args, **kwargs):
        self.configure(state="normal")
        super().insert(*args, **kwargs)
        self.configure(state="disabled")

    def delete(self, *args, **kwargs):
        self.configure(state="normal")
        super().delete(*args, **kwargs)
        self.configure(state="disabled")

    def show(self):
        self.window.deiconify()

    def hide(self):
        self.window.withdraw()


class _DerivedDefaultDict(defaultdict):
    def __missing__(self, key):
        value = self.default_factory(key)
        self[key] = value
        return value


class WgtStateChanger:
    def __init__(self, root):
        self.root = root
        self.bars_genres = tuple("dip rot vosc vrot losc lrot raman1 roa1".split())
        self.energies_genres = tuple("zpe ent ten gib scf".split())
        self._dependencies = {}
        self._genres = _DerivedDefaultDict(
            lambda g, r=self.root: r.tesliper.conformers.has_genre(g)
        )
        self._standards = {
            "energies": self._energies,
            "bars": self._bars,
            "tesliper": self._tesliper,
        }

    def _energies(self):
        return any(self._genres[g] for g in self.energies_genres)

    def _bars(self):
        return any(self._genres[g] for g in self.bars_genres)

    def _tesliper(self):
        return bool(self.root.tesliper.conformers)

    def register(
        self,
        widgets,
        dependencies=None,
        needs_all_genres=None,
        needs_any_genre=None,
        key=None,
    ):
        if isinstance(dependencies, str):
            dependencies = [dependencies]
        elif dependencies is None:
            dependencies = []
        for dep in dependencies:
            if dep not in self._standards:
                raise ValueError(f"Unknown dependency: {dep}.")
        if not isinstance(widgets, Iterable):
            widgets = [widgets]
        for wgt in widgets:

            def _conditions(
                obj=self,
                deps=dependencies,
                allg=needs_all_genres,
                anyg=needs_any_genre,
                key=key,
            ):
                dep_ = all(obj._standards[d]() for d in deps)
                all_ = all(obj._genres[g] for g in allg) if allg else True
                any_ = any(obj._genres[g] for g in anyg) if anyg else True
                key_ = key() if key else True
                return all([dep_, all_, any_, key_])

            self._dependencies[wgt] = _conditions

    def set_states(self):
        self._genres.clear()
        for widget, condition in self._dependencies.items():
            changer = self.enable if condition() else self.disable
            changer(widget)

    @staticmethod
    def enable(widget):
        if isinstance(widget, ttk.Combobox):
            widget.configure(state="readonly")
        else:
            widget.configure(state="normal")

    @staticmethod
    def disable(widget):
        widget.configure(state="disabled")


class FeedbackThread(Thread):
    def __init__(self, gui, progbar_msg, target, args, kwargs):
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.progbar_msg = progbar_msg
        self.gui = gui
        self.queue = queue.Queue()
        super().__init__(daemon=True)

    def run(self):
        self.exc = None
        self.gui.progtext.set(self.progbar_msg)
        self.gui.progbar.configure(mode="indeterminate")
        self.gui.progbar.start()
        try:
            return_value = self.target(*self.args, **self.kwargs)
            self.queue.put(return_value)
        except BaseException as exc:
            self.exc = exc
        self.gui.progbar.stop()
        self.gui.progbar.configure(mode="determinate")
        self.gui.progtext.set("Idle.")
        if self.exc:
            logger.critical("Something unexpected happend.", exc_info=self.exc)
            return
            # raise self.exc
        else:
            return return_value


class ThreadedMethod:
    def __init__(self, progbar_msg):
        self.progbar_msg = progbar_msg

    def __call__(self, function):
        def wrapper(widget, *args, **kwargs):
            # widget is *self* from decorated method
            root = widget.winfo_toplevel()
            if root.thread.is_alive():
                msg = "Can't start {}, while {} is still running.".format(
                    function, root.thread.target
                )
                logger.info(msg)
                return  # log and do nothing
            else:
                root.thread = FeedbackThread(
                    root,
                    self.progbar_msg,
                    function,
                    [widget] + list(args),
                    kwargs,
                )
            root.thread.start()

        return wrapper



"""Functions that deal with spectra and spectral data."""
import logging as lgg
import math
from typing import Callable, Sequence, Tuple, Union

import numpy as np

# LOGGER
logger = lgg.getLogger(__name__)
logger.setLevel(lgg.DEBUG)


# TYPES
Number = Union[int, float]
Numbers = Sequence[Union[int, float]]
FittingFunctionType = Callable[[np.ndarray, np.ndarray, np.ndarray, float], np.ndarray]


# MODULE FUNCTIONS
def count_imaginary(frequencies: np.ndarray):
    """Finds number of imaginary frequencies of each conformer.

    Parameters
    ----------
    frequencies
        List of conformers' frequencies. Array with one dimension is interpreted
        as list of frequencies for single conformer.

    Returns
    -------
    numpy.ndarray
        Number of imaginary frequencies of each conformer.

    Raises
    ------
    ValueError
        If input array has more than 2 dimensions."""
    if frequencies.size == 0:
        return np.array([])
    elif frequencies.ndim < 2:
        return np.asarray((frequencies < 0).sum())
    elif frequencies.ndim == 2:
        return (frequencies < 0).sum(1)
    else:
        raise ValueError(
            f"Array with {frequencies.ndim} dimensions can't be interpreted "
            f"as a list of conformers' frequencies."
        )


def find_imaginary(frequencies: np.ndarray):
    """Finds all conformers with imaginary frequency values.

    Parameters
    ----------
    frequencies
        List of conformers' frequencies.

    Returns
    -------
    numpy.ndarray
        List of the indices of conformers with imaginary frequency values.

    Raises
    ------
    ValueError
        If input array has more than 2 dimensions."""
    try:
        imag = count_imaginary(frequencies)
    except ValueError as err:
        raise ValueError(err) from err
    return np.nonzero(imag)[0]


def gaussian(
    intensities: np.ndarray,
    frequencies: np.ndarray,
    abscissa: np.ndarray,
    width: Number,
) -> np.ndarray:
    """Gaussian fitting function for spectra calculation.

    Parameters
    ----------
    intensities
        Appropriate values extracted from gaussian output files.
    frequencies
        Frequencies extracted from gaussian output files.
    abscissa
        List of wavelength/wave number points on spectrum x axis.
    width
        Number representing half width of peak at 1/e its maximum height.

    Returns
    -------
    numpy.ndarray
        List of calculated intensity values.

    Raises
    ------
    ValueError
        If given width is not greater than zero.
        If *intensities* and *frequencies* are not of the sane shape."""
    if width <= 0:
        raise ValueError("Peak width must be a positive value!")
    if intensities.shape != frequencies.shape:
        raise ValueError("*intensities* and *frequencies* must be of same shape!")
    if abscissa.size == 0:
        return np.array([])
    sigm = width / 1.4142135623730951  # math.sqrt(2)
    denominator = sigm * 2.5066282746310002  # (2 * math.pi) ** 0.5
    it = np.nditer(
        [abscissa, None],
        flags=["buffered"],
        op_flags=[["readonly"], ["writeonly", "allocate", "no_broadcast"]],
        op_dtypes=[np.float64, np.float64],
    )
    for x, peaks in it:
        e = intensities * np.exp(-0.5 * ((x - frequencies) / sigm) ** 2)
        peaks[...] = (e / denominator).sum()
    return it.operands[1]


def lorentzian(
    intensities: np.ndarray,
    frequencies: np.ndarray,
    abscissa: np.ndarray,
    width: Number,
) -> np.ndarray:
    """Lorentzian fitting function for spectra calculation.

    Parameters
    ----------
    intensities
        Appropriate values extracted from gaussian output files.
    frequencies
        Frequencies extracted from gaussian output files.
    abscissa
        List of wavelength/wave number points on spectrum x axis.
    width
        Number representing half width of peak at half its maximum height.

    Returns
    -------
    numpy.ndarray
        List of calculated intensity values.

    Raises
    ------
    ValueError
        If given width is not greater than zero.
        If *intensities* and *frequencies* are not of the same shape."""
    if width <= 0:
        raise ValueError("Peak width must be a positive value!")
    if intensities.shape != frequencies.shape:
        raise ValueError("*intensities* and *frequencies* must be of same shape!")
    if abscissa.size == 0:
        return np.array([])
    hwhmsqrd = width ** 2
    hwhmoverpi = width / math.pi
    it = np.nditer(
        [abscissa, None],
        flags=["buffered"],
        op_flags=[["readonly"], ["writeonly", "allocate", "no_broadcast"]],
        op_dtypes=[np.float64, np.float64],
    )
    for x, val in it:
        s = intensities / ((frequencies - x) ** 2 + hwhmsqrd)
        s2 = (hwhmoverpi * s).sum()
        val[...] = s2
    return it.operands[1]


def calculate_spectra(
    frequencies: np.ndarray,
    intensities: np.ndarray,
    abscissa: np.ndarray,
    width: Number,
    fitting: FittingFunctionType,
):
    """Calculates spectrum for each individual conformer.

    Parameters
    ----------
    frequencies
        List of conformers' frequencies in cm^(-1). Should be of shape
        (number _of_conformers, number_of_frequencies).
    intensities
        List of calculated signal intensities for each conformer. Should be
        of same shape as frequencies.
    abscissa
        List of points on x axis in output spectrum in cm^(-1).
    width : int or float
        Number representing peak width in cm^(-1), used by fitting function.
    fitting : function
        Function, which takes intensities, frequencies, abscissa, hwhm as
        parameters and returns numpy.array of calculated spectrum points.

    Returns
    -------
    numpy.ndarray
        Array of intensity values for each conformer.

    Raises
    ------
    ValueError
        If given width is not greater than zero.
        If *intensities* and *frequencies* are not of the same shape."""
    if intensities.shape != frequencies.shape:
        raise ValueError("*intensities* and *frequencies* must be of same shape!")
    if not intensities.size:
        return np.zeros(0)  # return early to avoid (0, N) shape of output array
    spectra = np.zeros([len(frequencies), abscissa.shape[0]])  # template
    for inten, freq, spr in zip(intensities, frequencies, spectra):
        spr[...] = fitting(inten, freq, abscissa, width)
    return spectra


def calculate_average(
    values: Union[Numbers, np.ndarray], populations: Union[Numbers, np.ndarray]
) -> np.ndarray:
    """Calculates weighted average of *values*, where *populations* are used as
    weights.

    Parameters
    ----------
    values
        List of values for each conformer, should be of shape (N, M), where N is
        number of conformers and M is number of values.
    populations
        List of conformers' populations, should be of shape (N,) where N is
        number of conformers. Should add up to 1.

    Returns
    -------
    numpy.ndarray
        weighted arithmetic mean of values given.

    Raises
    ------
    ValueError
        If parameters of non-matching shape were given."""
    values = np.asanyarray(values)
    populations = np.asanyarray(populations)
    if not populations.size == values.shape[0]:
        raise ValueError(
            "Exactly one population value for each conformer must be provided."
        )
    if not values.size:
        return np.zeros([0])  # just return an empty array if *values* is empty
    popsum = populations.sum()
    if not np.isclose(popsum, 1):
        # normalize population data, if needed
        populations = populations / popsum
    # populations must be of same shape as values array
    # so we expand populations with appropriate number of dimensions
    shape = (-1,) + (1,) * (values.ndim - 1)
    popul = populations.reshape(*shape)
    return (values * popul).sum(0)


def idx_offset(a: Numbers, b: Numbers) -> int:
    """Calculate offset by which *b* should be shifted to best overlap with *a*.
    Both *a* and *b* should be sets of points, interpreted as spectral data. Returned
    offset is a number of data points, by which *b* should be moved relative to *a*,
    to get the best overlap of given spectra.

    Parameters
    ----------
    a
        *y* values of the first spectrum.
    b
        *y* values of the second spectrum.

    Returns
    -------
    int
        Offset, in number of data points, by which spectrum *b* should be shifted
        to best match spectrum *a*. Positive value means it should be shifted to the
        right and negative value means it should be shifted to the left of *a*.

    Notes
    -----
    The best overlap is found by means of cross-correlation of given spectra.
    """
    a, b = np.asanyarray(a), np.asanyarray(b)
    # normalize values to be zero centered to prevent influence of padding with zeros
    a = (a - a.mean()) / a.std()
    b = (b - b.mean()) / b.std()
    # calculate cross correlation array and find best overlap
    cross = np.correlate(a, b, mode="full")
    best = cross.argmax()
    return best - b.size + 1


def unify_abscissa(
    ax: Numbers, ay: Numbers, bx: Numbers, by: Numbers, upscale: bool = True
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Interpolate one of the given spectra to have the same points density as the
    other given spectrum.

    Which spectra should be interpolated is determined based on the density of points
    of both spectra, by default more loosely spaced spectrum is interpolated to match
    spacing of the other spectrum. This may be changed by passing ``upscale=False``
    to the function call.

    Parameters
    ----------
    ax
        Abscissa of the first spectrum.
    ay
        Values of the first spectrum.
    bx
        Abscissa of the second spectrum.
    by
        Values of the second spectrum.
    upscale
        If interpolation should be done on more loosely spaced spectrum (default).
        When set to False, spectrum with lower resolution will be treated as reference.

    Returns
    -------
    tuple of np.arrays of numbers
        Spectra, one unchanged and one interpolated, as a tuple of numpy arrays
        of x and y values. I.e. ``tuple(ax, ay, new_bx, new_by)`` or
        ``tuple(new_ax, new_ay, bx, by)``, depending on values of *upscale* parameter.
    """
    ax, ay, bx, by = (
        np.asanyarray(ax),
        np.asanyarray(ay),
        np.asanyarray(bx),
        np.asanyarray(by),
    )
    ad, bd = ax[1] - ax[0], bx[1] - bx[0]  # we assume both have steady step
    if ad == bd:
        return ax, ay, bx, by  # no need to do anything
    elif (np.abs(ad) < np.abs(bd)) ^ upscale:  # xor on booleans
        # *ad* is smaller than *bd*, but we don't want to upscale or vice-versa
        nbx, nby, nax, nay = unify_abscissa(bx, by, ax, ay, upscale)  # swap spectra
        # but return in the same order as given in parameters
    else:
        step = np.abs(ad) * np.sign(bd)  # ad is new step, but sign from bd
        nbx = np.arange(bx[0], bx[-1], step)  # new abscissa
        nby = np.interp(nbx, bx, by)  # interpolate values on new abscissa
        nax, nay = ax, ay  # the other spectrum stays unchanged
    return nax, nay, nbx, nby


def find_offset(
    ax: Numbers, ay: Numbers, bx: Numbers, by: Numbers, upscale: bool = True
) -> float:
    """Finds value, by which the spectrum should be shifted along x-axis to best
    overlap with the first spectrum. If resolution of spectra is not identical,
    one of them will be interpolated to match resolution of the other one. By default
    interpolation is done on the lower-resolution spectra. This can be changed
    by passing ``upscale = False`` to function call.

    Parameters
    ----------
    ax
        Abscissa of the first spectrum.
    ay
        Values of the first spectrum.
    bx
        Abscissa of the second spectrum.
    by
        Values of the second spectrum.
    upscale
        If interpolation should be done on more loosely spaced spectrum (default).
        When set to False, spectrum with lower resolution will be treated as reference
        for density of data points.

    Returns
    -------
    float
        Value, by which second spectrum should be shifted, in appropriate units.
    """
    ax, ay, bx, by = unify_abscissa(ax, ay, bx, by, upscale=upscale)
    shift = idx_offset(ay, by)
    if shift < 0:
        offset = ax[0] - bx[abs(shift)]
    else:
        offset = ax[shift] - bx[0]
    return offset


def find_scaling(a: Numbers, b: Numbers) -> float:
    """Find factor by which values *b* should be scaled to best match values *a*.

    Parameters
    ----------
    a
        *y* values of the first spectrum.
    b
        *y* values of the second spectrum.

    Returns
    -------
    float
        Scaling factor for *b* values.

    Notes
    -----
    If scaling factor cannot be reasonably given, i.e. when *b* is an empty list or list
    of zeros or NaNs, ``1.0`` is returned. Values lower than 1% of maximum are ignored.
    """
    a, b = np.abs(a), np.abs(b)
    try:
        amax, bmax = max(a), max(b)
    except ValueError:
        return 1.0
    scaling = np.mean(a[a >= 0.01 * amax]) / np.mean(b[b >= 0.01 * bmax])
    scaling = 1.0 if np.isnan(scaling) else scaling
    return scaling


# genre: {target: converter}
_converters = {
    "freq": {"wavelen": lambda v: 1e7 / v, "ex_en": lambda v: v / 8065.544},
    "wavelen": {"freq": lambda v: 1e7 / v, "ex_en": lambda v: 1239.8 / v},
    "ex_en": {"freq": lambda v: v * 8065.544, "wavelen": lambda v: 1239.8 / v},
}


def convert_band(
    value: Union[float, np.ndarray], from_genre: str, to_genre: str
) -> Union[float, np.ndarray]:
    """Convert one representation of band to another.

    Parameters
    ----------
    value
        Value(s) to convert.
    from_genre
        Genre specifying a representation of band of input data.
        Should be one of: 'freq', 'wavelen', 'ex_en'.
    to_genre
        Genre specifying a representation of band, to which you want to convert.
        Should be one of: 'freq', 'wavelen', 'ex_en'.

    Returns
    -------
    float or np.ndarray
        Requested representation of bands. If *from_genre* is same as *to_genre*,
        then simply *value* is returned.
    """
    if from_genre == to_genre:
        return value
    try:
        converter = _converters[from_genre][to_genre]
    except KeyError as error:
        raise ValueError(
            f"Unsupported conversion: from '{from_genre}' to '{to_genre}'. "
            "Genres available for conversion are: 'freq', 'wavelen', 'ex_en'."
        ) from error
    return converter(value)



"""Export of Gaussian input files (.gjf) for setting up new calculation step."""
import logging
from itertools import cycle
from pathlib import Path
from string import Template
from typing import Dict, Iterable, List, Optional, Sequence, TextIO, Union

from ..datawork.atoms import symbol_of_element
from ..glassware import Geometry, IntegerArray
from .writer_base import WriterBase

# LOGGER
logger = logging.getLogger(__name__)


# FUNCTIONS
def _format_coordinates(coords: Sequence[Sequence[float]], atoms: Sequence[int]):
    for a, (x, y, z) in zip(atoms, coords):
        a = symbol_of_element(a)
        yield f" {a: <2} {x: > 12.8f} {y: > 12.8f} {z: > 12.8f}"


# CLASSES
class GjfWriter(WriterBase):
    """Generates Gaussian input files for each conformer given."""

    extension = "gjf"
    _link0_commands = {
        "Mem",  # str specifying required memory
        "Chk",  # str with file path
        "OldChk",  # str with file path
        "SChk",  # str with file path
        "RWF",  # str with file path
        "OldMatrix",  # str with file path
        "OldRawMatrix",  # str with file path
        "Int",  # str with spec
        "D2E",  # str with spec
        "KJob",  # str with link number and, optionally, space-separated number
        "Save",  # boolean
        "ErrorSave",  # boolean
        "NoSave",  # boolean, same as ErrorSave
        "Subst",  # str with link number and space-separated file path
    }
    _link0_commands = {k.lower(): k for k in _link0_commands}
    _parametrized = {
        "chk",
        "oldchk",
        "schk",
        "rwf",
        "oldmatrix",
        "oldrawmatrix",
        "subst",
    }
    empty_lines_at_end = 2

    def __init__(
        self,
        destination: Union[str, Path],
        mode: str = "x",
        link0: Optional[Dict[str, Union[str, bool]]] = None,
        route: str = "",
        comment: str = "No information provided.",
        post_spec: str = "",
    ):
        """
        Parameters
        ----------
        destination: str or pathlib.Path
            Directory, to which generated files should be written.
        mode: str, optional
            Specifies how writing to file should be handled. Should be one of
            characters: "a" (append to existing file); "x" (only write if file doesn't
            exist yet); or "w" (overwrite file if it already exists). Defaults to "x".
        link0 : Dict[str, Union[str, bool]], optional
            Link0 commands that should be included in generated files, as a dictionary
            of {"command": "value"}. Refer to :attr:`link0` for more information. If
            omitted, no link0 commands are added.
        route : str
            Calculation directives for Gaussan, refer to the Gaussian documentation for
            information on how to construct the calculations route.
        comment : str, optional
            Additional text, describing the calculations, by default "No information
            provided."
        post_spec : str, optional
            Additional specification written after the molecule specification, written
            to generated files as provided by the user (you need to take care of line
            breaks). If omitted, no additional specification is added.
        """
        super().__init__(destination=destination, mode=mode)
        self.link0 = link0 or {}
        self.route = route
        self.comment = comment
        self.post_spec = post_spec

    def geometry(
        self,
        geometry: Geometry,
        charge: Union[IntegerArray, Sequence[int], int, None] = None,
        multiplicity: Union[IntegerArray, Sequence[int], int, None] = None,
        name_template: Union[str, Template] = "${conf}.${ext}",
    ):
        """Write given conformers' geometries to multiple Gaussian input files.

        Parameters
        ----------
        geometry : Geometry
            :class:`.Geometry` object containing data for each confomer that should be
            exported as Gaussian input file.
        charge : Union[IntegerArray, Sequence[int], int, None], optional
            Molecule's charge for each conformer. May be a sequence of values or one
            value that will be repeated for each conformer. By default 0 for each.
        multiplicity : Union[IntegerArray, Sequence[int], int, None], optional
            Molecule's multiplicity for each conformer. May be a sequence of values or
            one value that will be repeated for each conformer. By default 1 for each.
        name_template : Union[str, Template], optional
            Template that will be used to generate filenames, by default
            "${conf}.${ext}". Refer to :meth:`.make_name` documentation for details on
            supported placeholders.
        """
        geom = geometry.values
        atoms = cycle(geometry.atoms)
        try:
            char = charge.values
        except AttributeError:
            char = (0,) if charge is None else charge
            char = [char] if not isinstance(char, Iterable) else char
        char = cycle(char)
        try:
            mult = multiplicity.values
        except AttributeError:
            mult = (1,) if multiplicity is None else multiplicity
            mult = [mult] if not isinstance(mult, Iterable) else mult
        mult = cycle(mult)
        template_params = {"genre": geometry.genre, "cat": "geometry"}
        for *params, handle in zip(
            geom,
            atoms,
            char,
            mult,
            self._iter_handles(geometry.filenames, name_template, template_params),
        ):
            self._write_conformer(handle, *params, template_params)

    def _write_conformer(
        self,
        file: TextIO,
        g: Sequence[Sequence[float]],
        a: Sequence[int],
        c: int,
        m: int,
        template_params: dict,
    ):
        for key, value in self.link0.items():
            key = key.lower()
            if key in self._parametrized:
                # template_params are updated bt _iter_handles
                # so we can simply reuse it
                value = self.make_name(template=value, **template_params)
            if "save" in key and value:
                file.write(f"%{self._link0_commands[key]}\n")
            else:
                file.write(f"%{self._link0_commands[key]}={value}\n")
        file.write(self.route)
        file.write("\n" * 2)
        file.write(self.comment)
        file.write("\n" * 2)
        file.write(f"{c} {m}")
        for line in _format_coordinates(g, a):
            file.write("\n" + line)
        if self.post_spec:
            file.write("\n\n")
            file.write(self.post_spec)
        file.write("\n" * self.empty_lines_at_end)

    @property
    def link0(self) -> Dict[str, Union[str, bool]]:
        """
        Link0 commands, in a form of ``{"command": "value"}``, that will be placed in
        the beginning of each Gaussian input file created. If anny *command* is an
        unknown keword, an exception will be raised. Accepted *command* keywords are as
        follows:

        :Mem: str specifying required memory
        :Chk: str with file path
        :OldChk: str with file path
        :SChk: str with file path
        :RWF: str with file path
        :OldMatrix: str with file path
        :OldRawMatrix: str with file path
        :Int: str with spec
        :D2E: str with spec
        :KJob: str with link number and, optionally, space-separated number
        :Save: boolean
        :ErrorSave: boolean
        :NoSave: boolean, same as ErrorSave
        :Subst: str with link number and space-separated file path

        Commands that provide a file path as a value may be parametrized for each
        conformer. You can put a placeholder inside a given string path, that will be
        parametrized when writing to file. See :meth:`.make_name` to see available
        placeholders. You may use any of values listed there, however ``${conf}`` and
        ``${num}`` will probably be the most useful.
        """
        return self._link0

    @link0.setter
    def link0(self, commands: Dict[str, Union[str, bool]]):
        unknown = {k for k in commands if k.lower() not in self._link0_commands}
        if unknown:
            raise ValueError(f"Unknown link 0 commands provided: {', '.join(unknown)}.")
        self._link0 = {k: v for k, v in commands.items() if v}

    @property
    def route(self) -> str:
        """Also known as *# lines*, specifies desired calculation type, model chemistry,
        and other options for Gaussian. If pound sign is missing, it is added in the
        beginning. For supported keywords and syntax refer to the Gaussian's
        documentation.
        """
        return " ".join(self._route)

    @route.setter
    def route(self, commands: str):
        try:
            commands_ = commands.split()
        except AttributeError as error:
            raise TypeError("Expected object of type str.") from error
        length = len(commands_)
        if not length:
            commands_ = ["#"]
        else:
            first = commands_[0]
            if not first.startswith("#"):
                commands_ = ["#"] + commands_
        self._route = commands_



from unittest import mock
import pytest
from hypothesis import assume, given, strategies as st

from tesliper.extraction.parameters_parser import (
    ParametersParser,
    quantity,
    fitting,
    ParsingError,
)


@given(
    s1=st.text(alphabet=st.characters(blacklist_categories=("Cs", "Nd", "No"))),
    s2=st.text().filter(lambda x: not x or not x[0].isnumeric()),
    f=st.floats(allow_nan=False, allow_infinity=False),
)
def test_quantity(s1, s2, f):
    assume(not s1.endswith((".", "-")))
    test_string = f"{s1}{f}{s2}"
    assert f == quantity(test_string), test_string


@given(s=st.text(alphabet=st.characters(blacklist_categories=("Cs", "Nd", "No"))))
def test_quantity_raises(s):
    with pytest.raises(ParsingError):
        quantity(s)


@given(
    s1=st.text(),
    s2=st.text().filter(lambda x: not x or not x[0].isnumeric()),
    f=st.one_of(st.just("gaussian"), st.just("lorentzian")),
)
def test_fitting(s1, s2, f):
    test_string = f"{s1}{f}{s2}"
    assert callable(fitting(test_string))


@given(
    s=st.text().filter(lambda x: all(f not in x for f in ["gaussian", "lorentzian"]))
)
def test_fitting_raises(s):
    with pytest.raises(ParsingError):
        fitting(s)


@pytest.fixture
def get_config(tmp_path):
    def write_config(config):
        file = tmp_path / "par.txt"
        with file.open("w") as handle:
            handle.write(config)
        return file

    return write_config


@pytest.fixture
def stdconfig(get_config):
    return (
        "[PARAMETERS]\n"
        "width = 1\n"
        "start = 0\n"
        "stop = 100\n"
        "step = 5\n"
        "fitting = gaussian\n"
    )


def test_parser(stdconfig, get_config):
    file = get_config(stdconfig)
    p = ParametersParser()
    data = p.parse(file)
    assert data.keys() == p._transformers.keys()
    assert callable(data["fitting"])
    assert all(isinstance(data[k], float) for k in ("width", "start", "stop", "step"))
    assert data == p.parameters


def test_parser_aliases(get_config):
    file = get_config(
        "[PARAMETERS]\n"
        "Half width of band in half height = 1\n"
        "Start Range = 0\n"
        "Stop range = 100\n"
        "Step = 5\n"
        "Fitting Function = Gaussian\n"
    )
    p = ParametersParser()
    data = p.parse(file)
    assert data.keys() == p._transformers.keys()


def test_parser_missing_param(stdconfig, get_config):
    file = get_config("[PARAMETERS]\nwidth = 1\n")
    p = ParametersParser()
    data = p.parse(file)
    assert list(data) == ["width"]


def test_parser_invalid_param(stdconfig, get_config):
    file = get_config("[PARAMETERS]\nwidth = one\n")
    p = ParametersParser()
    data = p.parse(file)
    assert list(data) == []


def test_parser_unknown_param(stdconfig, get_config):
    file = get_config("[PARAMETERS]\nother = one\n")
    p = ParametersParser()
    data = p.parse(file)
    assert list(data) == ["other"]


def test_parser_empty(stdconfig, get_config):
    file = get_config("[PARAMETERS]\n")
    p = ParametersParser()
    data = p.parse(file)
    assert list(data) == []


def test_parser_missing_header(stdconfig, get_config):
    file = get_config("\n".join(stdconfig.split("\n")[1:]))
    p = ParametersParser()
    data = p.parse(file)
    assert data.keys() == p._transformers.keys()


def test_parser_different_header(stdconfig, get_config):
    config = ["[SPAM]\n"] + stdconfig.split("\n")[1:]
    file = get_config("\n".join(config))
    p = ParametersParser()
    data = p.parse(file)
    assert data.keys() == p._transformers.keys()


def test_parser_two_sections(stdconfig, get_config):
    config = stdconfig + "\n[OTHER]\nfoo = bar\n"
    file = get_config(config)
    p = ParametersParser()
    with pytest.raises(ParsingError):
        p.parse(file)



from .checktree import BoxVar, CheckTree, ConformersOverview, EnergiesView
from .collapsible_pane import CollapsiblePane
from .controls import CalculateSpectra, EnergiesChoice, FilterRMSD
from .helpers import (
    MaxLevelFilter,
    PopupHandler,
    ReadOnlyText,
    ShortExcFormatter,
    TextHandler,
    ThreadedMethod,
    WgtStateChanger,
)
from .label_separator import LabelSeparator
from .popups import ExportPopup
from .scroll import AutoScrollbar, ScrollableFrame
from .views import SpectraView
